#!/usr/bin/python
#
# Copyright (c) 2016 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import hashlib
import itertools
import math
import os
import random

import pysam

import paleomix.common.fileutils as fileutils
import paleomix.common.rtools as rtools
import paleomix.common.versions as versions
import paleomix.tools.factory as factory

from paleomix.atomiccmd.builder import AtomicCmdBuilder
from paleomix.atomiccmd.command import AtomicCmd
from paleomix.atomiccmd.sets import SequentialCmds
from paleomix.node import CommandNode, Node, NodeError

from paleomix.tools.zonkey.common import \
    RSCRIPT_VERSION, \
    contig_name_to_plink_name, \
    read_summary


ADMIXTURE_VERSION = versions.Requirement(call=("admixture", "--version"),
                                         search="(\d+)\.(\d+)",
                                         checks=versions.GE(1, 3))

PLINK_VERSION = versions.Requirement(call=("plink", "--noweb", "--help",
                                           "--out", "/tmp/plink"),
                                     search="v(\d+)\.(\d+)",
                                     checks=versions.GE(1, 7))

SMARTPCA_VERSION = versions.Requirement(call=("smartpca",),
                                        search="version: (\d+)",
                                        checks=versions.GE(13050))

TREEMIX_VERSION = versions.Requirement(call=("treemix",),
                                       search="TreeMix v. (\d+)\.(\d+)",
                                       checks=versions.GE(1, 12))


class BuildTPEDFilesNode(CommandNode):
    def __init__(self, output_root, table, bamfile, downsample,
                 dependencies=()):
        cmd = factory.new("zonkey_tped")
        cmd.set_option("--name", "Sample")
        cmd.set_option("--downsample", downsample)
        cmd.add_value("%(TEMP_DIR)s")
        cmd.add_value("%(IN_TABLE)s")
        cmd.add_value("%(IN_BAM)s")

        if not downsample:
            # Needed for random access (chromosomes are read 1 ... 31)
            cmd.set_kwargs(IN_BAI=fileutils.swap_ext(bamfile, ".bai"))

        cmd.set_kwargs(OUT_TFAM=os.path.join(output_root, "common.tfam"),
                       OUT_SUMMARY=os.path.join(output_root, "common.summary"),
                       OUT_TPED_INCL_TS=os.path.join(output_root,
                                                     "incl_ts.tped"),
                       OUT_TPED_EXCL_TS=os.path.join(output_root,
                                                     "excl_ts.tped"),
                       IN_TABLE=table,
                       IN_BAM=bamfile)

        CommandNode.__init__(self,
                             description="<BuildTPEDFiles -> %r>"
                             % (os.path.join(output_root, '*'),),
                             command=cmd.finalize(),
                             dependencies=dependencies)


class BuildBEDFilesNode(CommandNode):
    def __init__(self, output_prefix, tfam, tped,
                 indep_filter=None, indep_parameters=None,
                 plink_parameters=None,
                 dependencies=()):
        temp_prefix = os.path.basename(output_prefix)

        plink_cmd = ["plink", "--make-bed", "--noweb",
                     "--tped", "%(IN_TPED)s",
                     "--tfam", "%(IN_TFAM)s",
                     "--out", "%(TEMP_OUT_PREFIX)s"]

        plink_cmd.extend(self._parse_parameters(plink_parameters))

        command = AtomicCmd(plink_cmd,
                            IN_TPED=tped,
                            IN_TFAM=tfam,
                            TEMP_OUT_PREFIX=temp_prefix,
                            OUT_BED=output_prefix + ".bed",
                            OUT_BIM=output_prefix + ".bim",
                            OUT_FAM=output_prefix + ".fam",
                            OUT_LOG=output_prefix + ".log",
                            TEMP_OUT_NOSEX=temp_prefix + ".nosex",
                            TEMP_OUT_NOF=temp_prefix + ".nof",
                            CHECK_VERSION=PLINK_VERSION,
                            set_cwd=True)

        CommandNode.__init__(self,
                             description="<BuildBEDFiles -> '%s.*'>"
                             % (output_prefix,),
                             command=command,
                             dependencies=dependencies)

    @classmethod
    def _parse_parameters(cls, parameters):
        return parameters.split()


class BuildFilteredBEDFilesNode(CommandNode):
    def __init__(self, output_prefix, tfam, tped,
                 indep_filter=None, indep_parameters=None,
                 plink_parameters=None,
                 dependencies=()):

        assert indep_filter in ('indep',
                                'indep-pairphase',
                                'indep-pairwise'), indep_filter
        assert len(indep_parameters) == 3, indep_parameters

        parameters = self._parse_parameters(plink_parameters)

        plink_cmd = ["plink", "--noweb",
                     "--tped", "%(IN_TPED)s",
                     "--tfam", "%(IN_TFAM)s",
                     "--out", "%(TEMP_OUT_PREFIX)s",
                     '--' + indep_filter]
        plink_cmd.extend(indep_parameters)
        plink_cmd.extend(parameters)

        cmd_indep = AtomicCmd(plink_cmd,
                              IN_TFAM=tfam,
                              IN_TPED=tped,
                              TEMP_OUT_PREFIX="indep",
                              TEMP_OUT_LOG="indep.log",
                              TEMP_OUT_NOSEX="indep.nosex",
                              TEMP_OUT_PRUNE_IN="indep.prune.in",
                              TEMP_OUT_PRUNE_OUT="indep.prune.out",
                              set_cwd=True)

        basename = os.path.basename(output_prefix)
        cmd_filter = AtomicCmd(["plink", "--noweb", "--make-bed",
                                "--tped", "%(IN_TPED)s",
                                "--tfam", "%(IN_TFAM)s",
                                "--extract", "%(TEMP_IN_PRUNE)s",
                                "--out", "%(TEMP_OUT_PREFIX)s"] +
                               parameters,
                               IN_TFAM=tfam,
                               IN_TPED=tped,
                               TEMP_OUT_PREFIX=basename,
                               TEMP_IN_PRUNE="indep.prune.in",
                               TEMP_OUT_NOSEX=basename + ".nosex",
                               TEMP_OUT_LOG=basename + ".log",
                               OUT_LOG=output_prefix + ".log",
                               OUT_BED=output_prefix + ".bed",
                               OUT_BIM=output_prefix + ".bim",
                               OUT_FAM=output_prefix + ".fam",
                               set_cwd=True)

        CommandNode.__init__(self,
                             description="<BuildFilteredBEDFiles -> '%s.*'>"
                             % (output_prefix,),
                             command=SequentialCmds((cmd_indep, cmd_filter)),
                             dependencies=dependencies)

    @classmethod
    def _parse_parameters(cls, parameters):
        return parameters.split()


class AdmixtureNode(CommandNode):
    def __init__(self, input_file, k_groups, output_root,
                 samples=None, dependencies=()):
        self._samples = samples
        self._input_file = input_file
        self._k_groups = k_groups

        group_key = "Group(%i)" % (self._k_groups,)
        self._supervised = samples and any((row[group_key] != '-')
                                           for row in samples.itervalues())

        assert k_groups in (2, 3), k_groups
        prefix = os.path.splitext(os.path.basename(input_file))[0]
        output_prefix = os.path.join(output_root,
                                     "%s.%i" % (prefix, k_groups))

        cmd = AtomicCmdBuilder("admixture",
                               IN_FILE_BED=input_file,
                               IN_FILE_BIM=fileutils.swap_ext(input_file,
                                                              ".bim"),
                               IN_FILE_FAM=fileutils.swap_ext(input_file,
                                                              ".fam"),

                               TEMP_OUT_FILE_BED=prefix + ".bed",
                               TEMP_OUT_FILE_BIM=prefix + ".bim",
                               TEMP_OUT_FILE_FAM=prefix + ".fam",
                               TEMP_OUT_FILE_POP=prefix + ".pop",

                               OUT_P=output_prefix + ".P",
                               OUT_Q=output_prefix + ".Q",
                               OUT_STDOUT=output_prefix + ".log",

                               CHECK_VERSION=ADMIXTURE_VERSION,
                               set_cwd=True)

        cmd.set_option("-s", random.randint(0, 2 ** 16 - 1))

        if self._supervised:
            cmd.set_option("--supervised")

        cmd.add_value("%(TEMP_OUT_FILE_BED)s")
        cmd.add_value(int(k_groups))

        CommandNode.__init__(self,
                             description="<Admixture -> '%s.*''>"
                             % (output_prefix,),
                             command=cmd.finalize(),
                             dependencies=dependencies)

    def _setup(self, config, temp):
        CommandNode._setup(self, config, temp)

        input_files = [
            self._input_file,
            fileutils.swap_ext(self._input_file, ".bim"),
            fileutils.swap_ext(self._input_file, ".fam"),
        ]

        for filename in input_files:
            basename = os.path.basename(filename)
            os.symlink(os.path.abspath(filename), os.path.join(temp, basename))

        if self._supervised:
            fam_filename = fileutils.swap_ext(self._input_file, ".fam")

            pop_filename = fileutils.swap_ext(fam_filename, ".pop")
            pop_filename = fileutils.reroot_path(temp, pop_filename)

            key = "Group(%i)" % (self._k_groups,)
            with open(fam_filename) as fam_handle:
                with open(pop_filename, "w") as pop_handle:
                    for line in fam_handle:
                        sample, _ = line.split(None, 1)
                        group = self._samples.get(sample, {}).get(key, "-")

                        pop_handle.write("%s\n" % (group,))


class SelectBestAdmixtureNode(Node):
    def __init__(self, replicates, output_root, dependencies=()):
        replicates = tuple(replicates)
        if not replicates:
            raise ValueError("No replicates passed to SelectBestAdmixture")

        input_files = []
        ref_filenames = None
        for node in replicates:
            filenames = frozenset(os.path.basename(filename)
                                  for filename in node.output_files)

            if ref_filenames is None:
                ref_filenames = filenames
            elif ref_filenames != filenames:
                raise RuntimeError("Node %r does not contain expected files, "
                                   "%r, vs %r" % (node, ref_filenames,
                                                  filenames))

            input_files.extend(node.output_files)

        output_files = [os.path.join(output_root, filename)
                        for filename in ref_filenames]

        self._ref_filenames = ref_filenames
        self._files = tuple(node.output_files for node in replicates)
        self._output_root = output_root

        Node.__init__(self,
                      description="<SelectBestAdmixture -> %r>"
                      % (output_root,),
                      input_files=input_files,
                      output_files=output_files,
                      dependencies=tuple(dependencies) + tuple(replicates))

    def _run(self, config, temp):
        likelihoods = []
        for fileset in self._files:
            for filename in fileset:
                if filename.endswith(".log"):
                    likelihoods.append((self._read_admixture_log(filename),
                                        fileset))
                    break
            else:
                raise NodeError("No log-file found in list of admixture "
                                "output-files: %r" % (fileset,))

        _, fileset = max(likelihoods)
        for src_filename in fileset:
            dst_filename = fileutils.reroot_path(self._output_root,
                                                 src_filename)
            fileutils.copy_file(src_filename, dst_filename)

    @classmethod
    def _read_admixture_log(cls, filename):
        with open(filename) as handle:
            for line in handle:
                if line.startswith("Loglikelihood:"):
                    return float(line.split()[1])

            raise NodeError("Could not find likelihood value in log-file %r; "
                            "looking for line starting with 'Loglikelihood:'"
                            % (filename,))


class AdmixturePlotNode(CommandNode):
    def __init__(self, input_file, output_prefix, order, samples,
                 dependencies=()):
        self._samples = samples
        self._order = tuple(order) + ("Sample",)

        script = rtools.rscript("zonkey", "admixture.r")

        cmd = AtomicCmd(("Rscript", script, "%(IN_FILE)s",
                         "%(TEMP_OUT_NAMES)s", "%(TEMP_OUT_PREFIX)s"),
                        AUX_RSCRIPT=script,
                        IN_FILE=input_file,
                        IN_SAMPLES=samples,
                        OUT_PDF=output_prefix + ".pdf",
                        OUT_PNG=output_prefix + ".png",
                        TEMP_OUT_NAMES="samples.txt",
                        TEMP_OUT_PREFIX=os.path.basename(output_prefix),
                        CHECK_R=RSCRIPT_VERSION,
                        CHECK_R_GGPLOT2=rtools.requirement("ggplot2"),
                        CHECK_R_RESHAPE2=rtools.requirement("reshape2"),
                        set_cwd=True)

        CommandNode.__init__(self,
                             description="<AdmixturePlot -> '%s.*'>"
                             % (output_prefix,),
                             command=cmd,
                             dependencies=dependencies)

    def _setup(self, config, temp):
        samples = {}
        with open(self._samples) as handle:
            header = handle.readline().strip().split('\t')
            for line in handle:
                row = dict(zip(header, line.strip().split('\t')))
                samples[row["Name"]] = row

        with open(os.path.join(temp, "samples.txt"), "w") as handle:
            handle.write("{}\n".format("\t".join(header)))

            for name in self._order:
                row = samples[name]
                handle.write("{}\n".format("\t".join(row[key]
                                                     for key in header)))

        CommandNode._setup(self, config, temp)


class BuildFreqFilesNode(CommandNode):
    def __init__(self, input_prefix, output_prefix, tfam,
                 parameters=None, dependencies=()):
        basename = os.path.basename(output_prefix)

        plink_cmd = ["plink", "--freq", "--missing", "--noweb",
                     "--bfile", os.path.abspath(input_prefix),
                     "--within", "%(TEMP_OUT_CLUST)s",
                     "--out", "%(TEMP_OUT_PREFIX)s"]

        if parameters:
            plink_cmd.extend(parameters.split())

        plink = AtomicCmd(plink_cmd,
                          IN_BED=input_prefix + ".bed",
                          IN_BIM=input_prefix + ".bim",
                          IN_FAM=input_prefix + ".fam",
                          TEMP_OUT_CLUST="samples.clust",
                          TEMP_OUT_IMISS=basename + ".imiss",
                          TEMP_OUT_LMISS=basename + ".lmiss",
                          OUT_NOSEX=output_prefix + ".frq.strat.nosex",
                          OUT_LOG=output_prefix + ".frq.strat.log",
                          TEMP_OUT_PREFIX=basename,
                          CHECK_VERSION=PLINK_VERSION,
                          set_cwd=True)

        gzip = AtomicCmd(["gzip", "%(TEMP_IN_FREQ)s"],
                         TEMP_IN_FREQ=basename + ".frq.strat",
                         OUT_FREQ=output_prefix + ".frq.strat.gz")

        self._tfam = tfam
        self._basename = basename
        CommandNode.__init__(self,
                             description="<BuildFreqFiles -> '%s.*'"
                             % (output_prefix,),
                             command=SequentialCmds((plink, gzip)),
                             dependencies=dependencies)

    def _setup(self, config, temp):
        CommandNode._setup(self, config, temp)

        with open(self._tfam) as in_handle:
            samples = [line.split(None, 1)[0] for line in in_handle]

        with open(os.path.join(temp, "samples.clust"), "w") as handle:
            for sample in samples:
                handle.write("{0} {0} {0}\n".format(sample))

    def _teardown(self, config, temp):
        for ext in ("log", "nosex"):
            os.rename(os.path.join(temp, self._basename + "." + ext),
                      os.path.join(temp, self._basename + ".frq.strat." + ext))

        CommandNode._teardown(self, config, temp)


class FreqToTreemixNode(Node):
    def __init__(self, input_file, output_file, dependencies=()):
        Node.__init__(self,
                      description="<FreqToTreemix -> %r" % (output_file,),
                      input_files=(input_file,),
                      output_files=(output_file,),
                      dependencies=dependencies)

    def _run(self, _config, temp):
        input_file, = self.input_files
        output_file, = self.output_files
        temp_filename = os.path.basename(output_file)

        with open(os.path.join(temp, temp_filename), "w") as handle:
            header = None
            table = self._parse_freq_table(input_file)

            for _, rows in itertools.groupby(table, lambda row: row[:2]):
                if header is None:
                    rows = tuple(rows)  # Must not consume iterator
                    header = list(sorted(row[2] for row in rows))
                    handle.write("%s\n" % (" ".join(header)))

                result = []
                rows = dict((row[2], row) for row in rows)
                for sample in header:
                    _, _, _, mac, nchroms = rows[sample]
                    result.append("%s,%i" % (mac, int(nchroms) - int(mac)))

                handle.write("%s\n" % (" ".join(result),))

    def _teardown(self, config, temp):
        output_file, = self.output_files
        temp_file = os.path.join(temp, os.path.basename(output_file))

        fileutils.move_file(temp_file, output_file)
        Node._teardown(self, config, temp)

    @classmethod
    def _parse_freq_table(cls, filename):
        with fileutils.open_ro(filename) as handle:
            handle.readline()  # Skip header

            for line in handle:
                chrom, snp, clst, _, _, _, mac, nchroms = line.split()

                yield (chrom, snp, clst, int(mac), int(nchroms))


class TreemixNode(CommandNode):
    def __init__(self, data, input_file, output_prefix, m=0, k=100,
                 outgroup=(), dependencies=()):
        call = ["treemix",
                "-i", "%(IN_FILE)s",
                "-o", "%(TEMP_OUT_PREFIX)s",
                "-global",
                "-m", m]

        if outgroup:
            call.extend(("-root", ",".join(outgroup)))

        self._param_m = m
        self._param_outgroup = outgroup
        self._params_file = output_prefix + ".parameters.txt"

        if isinstance(k, int):
            call.extend(("-k", k))
            self._param_k = k
            self._k_file = self._k_field = None
        elif isinstance(k, tuple) and all(isinstance(v, str) for v in k):
            self._k_field, self._k_file = k
            self._genome_size = sum(value["Size"]
                                    for value in data.contigs.itervalues())
            self._snp_distance = data.settings["SNPDistance"]
        else:
            raise ValueError("k must be int or (key, path) in TreemixNode")

        self._parameters_hash \
            = "%s.%s" % (output_prefix,
                         hash_params(k=k, m=m, global_set=True,
                                     outgroup=tuple(sorted(outgroup))))

        cmd = AtomicCmd(call,
                        IN_FILE=input_file,
                        TEMP_OUT_PREFIX=os.path.basename(output_prefix),
                        OUT_FILE_COV=output_prefix + ".cov.gz",
                        OUT_FILE_COVSE=output_prefix + ".covse.gz",
                        OUT_FILE_EDGES=output_prefix + ".edges.gz",
                        OUT_FILE_LLIK=output_prefix + ".llik",
                        OUT_FILE_MODELCOV=output_prefix + ".modelcov.gz",
                        OUT_FILE_TREEOUT=output_prefix + ".treeout.gz",
                        OUT_FILE_VERTICES=output_prefix + ".vertices.gz",
                        OUT_FILE_PARAMS=self._params_file,
                        OUT_FILE_PARAMS_HASH=self._parameters_hash,
                        CHECK_VERSION=TREEMIX_VERSION,
                        set_cwd=True)

        CommandNode.__init__(self,
                             description="<Treemix -> '%s.*'>"
                             % (output_prefix,),
                             command=cmd,
                             dependencies=dependencies)

    def _setup(self, config, temp):
        if self._k_file is not None:
            stats = read_summary(self._k_file)
            n_sites = float(stats[self._k_field])
            k = max(1, int(math.ceil(self._snp_distance /
                                     (self._genome_size / n_sites))))

            self._param_k = k
            self._command._command.extend(("-k", str(k)))

        CommandNode._setup(self, config, temp)

    def _teardown(self, config, temp):
        with open(fileutils.reroot_path(temp, self._params_file), "w") as out:
            out.write("k: %i\n" % (self._param_k,))
            out.write("m: %i\n" % (self._param_m,))
            out.write("outgroup: %r\n" % (list(self._param_outgroup),))

        open(fileutils.reroot_path(temp, self._parameters_hash), "w").close()

        CommandNode._teardown(self, config, temp)


class PlotTreemixNode(CommandNode):
    def __init__(self, samples, prefix, output_prefix, dependencies=()):
        abs_prefix = os.path.abspath(prefix)
        basename = os.path.basename(output_prefix)

        # TreeMix plots with migration edges
        cmd_1 = self._plot_command(prefix, "plot_tree", abs_prefix,
                                   "%(IN_SAMPLES)s", "%(TEMP_OUT_PREFIX)s",
                                   IN_SAMPLES=samples,
                                   TEMP_OUT_PREFIX=basename + "_tree",
                                   OUT_PDF=output_prefix + "_tree.pdf",
                                   OUT_PNG=output_prefix + "_tree.png")

        # Heatmap showing TreeMix residuals
        cmd_2 = self._plot_command(prefix, "plot_residuals", abs_prefix,
                                   "%(IN_SAMPLES)s", "%(TEMP_OUT_PREFIX)s",
                                   IN_SAMPLES=samples,
                                   TEMP_OUT_PREFIX=basename + "_residuals",
                                   OUT_PDF=output_prefix + "_residuals.pdf",
                                   OUT_PNG=output_prefix + "_residuals.png")

        # Text file containing % of variance explained by model
        cmd_3 = self._plot_command(prefix, "variance", abs_prefix,
                                   "%(OUT_TXT)s",
                                   OUT_TXT=output_prefix + "_variance.txt")

        CommandNode.__init__(self,
                             description="<PlotTreemix -> '%s.*'>"
                             % (output_prefix,),
                             command=SequentialCmds((cmd_1, cmd_2, cmd_3)),
                             dependencies=dependencies)

    @classmethod
    def _plot_command(cls, input_prefix, *args, **kwargs):
        script = rtools.rscript("zonkey", "treemix.r")

        return AtomicCmd(("Rscript", script) + args,
                         AUX_RSCRIPT=script,
                         IN_FILE_COV=input_prefix + ".cov.gz",
                         IN_FILE_COVSE=input_prefix + ".covse.gz",
                         IN_FILE_EDGES=input_prefix + ".edges.gz",
                         IN_FILE_MODELCOV=input_prefix + ".modelcov.gz",
                         IN_FILE_VERTICES=input_prefix + ".vertices.gz",
                         CHECK_R=RSCRIPT_VERSION,
                         CHECK_R_BREW=rtools.requirement("RColorBrewer"),
                         set_cwd=True,
                         **kwargs)


class SmartPCANode(CommandNode):
    def __init__(self, input_prefix, output_prefix, nchroms, dependencies=()):
        self._input_prefix = input_prefix
        self._output_prefix = output_prefix
        self._nchroms = nchroms

        cmd = AtomicCmd(("smartpca", "-p", "%(TEMP_OUT_PARAMS)s"),
                        TEMP_OUT_PARAMS="parameters.txt",
                        IN_FILE_BED=input_prefix + ".bed",
                        IN_FILE_BIM=input_prefix + ".bim",
                        IN_FILE_FAM=input_prefix + ".fam",
                        OUT_STDOUT=output_prefix + ".log",
                        OUT_EVEC=output_prefix + ".evec",
                        OUT_EVAL=output_prefix + ".eval",
                        OUT_SNPS=output_prefix + ".deleted_snps",
                        CHECK_VERSION=SMARTPCA_VERSION,
                        set_cwd=True)

        CommandNode.__init__(self,
                             description="<SmartPCA -> '%s.*>"
                             % (output_prefix,),
                             command=cmd,
                             dependencies=dependencies)

    def _setup(self, config, temp):
        CommandNode._setup(self, config, temp)

        with open(os.path.join(temp, "parameters.txt"), "w") as handle:
            handle.write("""
genotypename:      {input_prefix}.bed
snpname:           {input_prefix}.bim
indivname:         {input_prefix}.fam
evecoutname:       {output_prefix}.evec
evaloutname:       {output_prefix}.eval
deletsnpoutname:   {output_prefix}.deleted_snps
altnormstyle:      NO
numoutevec:        5
familynames:       YES
numoutlieriter:    1
numchrom:          {nchroms}
numthreads:        1
""".format(input_prefix=os.path.abspath(self._input_prefix),
           output_prefix=os.path.basename(self._output_prefix),
           nchroms=self._nchroms))

    def _teardown(self, config, temp):
        # Ensure that this file exists even when no filtered SNPs
        deleted_snps = os.path.basename(self._output_prefix) + ".deleted_snps"
        open(os.path.join(temp, deleted_snps), "a").close()

        CommandNode._teardown(self, config, temp)


class PlotPCANode(CommandNode):
    def __init__(self, samples, prefix, output_prefix, dependencies=()):
        abs_prefix = os.path.abspath(prefix)

        script = rtools.rscript("zonkey", "pca.r")
        call = ["Rscript", script,
                abs_prefix, "%(IN_SAMPLES)s", "%(TEMP_OUT_PREFIX)s"]

        cmd = AtomicCmd(call,
                        AUX_SCRIPT=script,
                        IN_FILE_EVAL=prefix + ".eval",
                        IN_FILE_EVEC=prefix + ".evec",
                        IN_SAMPLES=samples,
                        TEMP_OUT_PREFIX=os.path.basename(output_prefix),
                        OUT_PDF=output_prefix + ".pdf",
                        OUT_PNG=output_prefix + ".png",
                        CHECK_R=RSCRIPT_VERSION,
                        CHECK_R_GGPLOT2=rtools.requirement("ggplot2"),
                        CHECK_R_LABELS=rtools.requirement("ggrepel"),
                        set_cwd=True)

        CommandNode.__init__(self,
                             description="<PlotPCA -> '%s.*'>"
                             % (output_prefix,),
                             command=cmd,
                             dependencies=dependencies)


class PlotCoverageNode(CommandNode):
    def __init__(self, contigs, input_file, output_prefix, dependencies=()):
        self._contigs = contigs
        self._input_file = input_file

        script = rtools.rscript("zonkey", "coverage.r")
        cmd = AtomicCmd(("Rscript", script,
                         "%(TEMP_OUT_TABLE)s", "%(TEMP_OUT_PREFIX)s"),
                        AUX_RSCRIPT=script,
                        IN_FILE=input_file,
                        TEMP_OUT_TABLE="contigs.table",
                        OUT_PDF=output_prefix + ".pdf",
                        OUT_PNG=output_prefix + ".png",
                        TEMP_OUT_PREFIX=os.path.basename(output_prefix),
                        CHECK_R=RSCRIPT_VERSION,
                        CHECK_R_GGPLOT2=rtools.requirement("ggplot2"),
                        set_cwd=True)

        CommandNode.__init__(self,
                             description="<CoveragePlot -> '%s.*'>"
                             % (output_prefix,),
                             command=cmd,
                             dependencies=dependencies)

    def _setup(self, config, temp):
        with open(os.path.join(temp, "contigs.table"), "w") as handle:
            handle.write("ID\tSize\tNs\tHits\n")

            # Workaround for pysam < 0.9 returning list, >= 0.9 returning str
            for line in "".join(pysam.idxstats(self._input_file)).split('\n'):
                line = line.strip()
                if not line:
                    continue

                name, size, hits, _ = line.split('\t')
                name = contig_name_to_plink_name(name)
                if name is None or not (name.isdigit() or name == 'X'):
                    continue
                elif name not in self._contigs:
                    # Excluding contigs is allowed
                    continue

                if int(size) != self._contigs[name]['Size']:
                    raise NodeError("Size mismatch between database and BAM; "
                                    "expected size %i, found %i for contig %r"
                                    % (int(size), self._contigs[name]['Size'],
                                       name))

                row = {
                    'ID': name,
                    'Size': self._contigs[name]['Size'],
                    'Ns': self._contigs[name]['Ns'],
                    'Hits': hits,
                }

                handle.write('{ID}\t{Size}\t{Ns}\t{Hits}\n'.format(**row))

        CommandNode._setup(self, config, temp)


def hash_params(*args, **kwargs):
    return hashlib.md5(repr([args, kwargs])).hexdigest()



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from nose.tools import \
     assert_equal, \
     assert_not_equal, \
     assert_raises

from paleomix.common.formats.newick import \
     Newick, \
     GraphError, \
     NewickError, \
     NewickParseError

from paleomix.common.testing import \
     assert_list_equal


###############################################################################
###############################################################################
# Constructor

def test_newick__constructor__name():
    node = Newick(name="AbC")
    assert_equal(node.name, "AbC")


def test_newick__constructor__children_set_in_internal_nodes():
    node = Newick(name="Leaf")
    top_node = Newick(children=[node])
    assert_equal(top_node.children, (node,))


def test_newick__constructor__children_not_set_in_leaf_nodes():
    node = Newick(name="Leaf")
    assert_equal(node.children, ())


def test_newick__constructor__is_leaf_true_for_leaf_nodes():
    node = Newick(name="Another Leaf")
    assert node.is_leaf


def test_newick__constructor__is_leaf_false_for_internal_nodes():
    node = Newick(name="Leaf")
    top_node = Newick(children=[node])
    assert not top_node.is_leaf


def test_newick__constuctor__leaf_nodes_must_have_name_or_length():
    assert_raises(NewickError, Newick, children=None)


def test_newick__constructor__internal_nodes_must_have_children():
    assert_raises(NewickError, Newick, children=[])


def test_newick__constructor__children_must_be_newick():
    assert_raises(TypeError, Newick, children=["A", "B"])


###############################################################################
###############################################################################
# get_leaf_nodes

def test_newick__get_leaf_nodes__leaf_returns_self():
    node = Newick(name="Leaf")
    assert_list_equal(node.get_leaf_nodes(), [node])


def test_newick__get_leaf_nodes__internal_node_returns_leaf_nodes():
    node_a = Newick(name="Leaf A")
    node_b = Newick(name="Leaf B")
    top_node = Newick(children=[node_a, node_b])
    assert_list_equal(top_node.get_leaf_nodes(), [node_a, node_b])


def test_newick__get_leaf_nodes__complex_case():
    node_a = Newick(name="Leaf A")
    node_b = Newick(name="Leaf B")
    node_c = Newick(name="Leaf C")
    sub_a = Newick(children=[node_b, node_c])
    top_node = Newick(children=[node_a, sub_a])
    assert_list_equal(top_node.get_leaf_nodes(), [node_a, node_b, node_c])


###############################################################################
###############################################################################
# get_leaf_nodes

def test_newick__get_leaf_names__leaf_returns_self():
    node = Newick(name="Leaf")
    assert_list_equal(node.get_leaf_names(), ["Leaf"])


def test_newick__get_leaf_names__internal_node_returns_leaf_nodes():
    node_a = Newick(name="Leaf A")
    node_b = Newick(name="Leaf B")
    top_node = Newick(children=[node_a, node_b])
    assert_list_equal(top_node.get_leaf_names(), ["Leaf A", "Leaf B"])


def test_newick__get_leaf_names__complex_case():
    node_a = Newick(name="Leaf A")
    node_b = Newick(name="Leaf B")
    node_c = Newick(name="Leaf C")
    sub_a = Newick(children=[node_b, node_c])
    top_node = Newick(children=[node_a, sub_a])
    assert_list_equal(top_node.get_leaf_names(), ["Leaf A", "Leaf B", "Leaf C"])


###############################################################################
###############################################################################
# reroot_on_taxa

def test_newick__reroot_on_taxa__single_taxa():
    source = Newick.from_string("((A,B),C);")
    expected = Newick.from_string("((B,C),A);")
    assert_equal(expected, source.reroot_on_taxa("A"))


def test_newick__reroot_on_taxa__single_taxa_with_branch_lengths():
    source = Newick.from_string("((A:4,B:3):2,C:1);")
    expected = Newick.from_string("((B:3,C:3.0):2.0,A:2.0);")
    assert_equal(expected, source.reroot_on_taxa("A"))


def test_newick__reroot_on_taxa__multiple_taxa__clade():
    source = Newick.from_string("((A,(B,C)),(D,E));")
    expected = Newick.from_string("(((D,E),A),(B,C));")
    assert_equal(expected, source.reroot_on_taxa(("B", "C")))


def test_newick__reroot_on_taxa__multiple_taxa__paraphylogeny():
    source = Newick.from_string("((B,C),((D,E),A));")
    expected = Newick.from_string("(((B,C),A),(D,E));")
    assert_equal(expected, source.reroot_on_taxa(("A", "C")))


def test_newick__reroot_on_taxa__no_taxa():
    source = Newick.from_string("((B,C),((D,E),A));")
    assert_raises(ValueError, source.reroot_on_taxa, ())


def test_newick__reroot_on_taxa__unknown_taxa():
    source = Newick.from_string("((B,C),((D,E),A));")
    assert_raises(ValueError, source.reroot_on_taxa, ("A", "Z"))


def test_newick__reroot_on_taxa__no_non_root_taxa():
    source = Newick.from_string("((B,C),((D,E),A));")
    assert_raises(ValueError, source.reroot_on_taxa, ("A", "B", "C", "D", "E"))


###############################################################################
###############################################################################
# reroot_on_midpoint

def test_newick__reroot_on_midpoint__single_node():
    source = Newick.from_string("(A:3.0);")
    expected = Newick.from_string("(A:3.0);")
    assert_equal(expected, source.reroot_on_midpoint())


def test_newick__reroot_on_midpoint__two_nodes():
    source = Newick.from_string("(A:3.0,B:8.0);")
    rerooted = source.reroot_on_midpoint()
    expected = Newick.from_string("(A:5.5,B:5.5);")
    assert_equal(expected, rerooted)


def test_newick__reroot_on_midpoint__two_clades():
    source = Newick.from_string("((A:7,B:2):1,(C:1,D:0.5):2);")
    rerooted = source.reroot_on_midpoint()
    expected = Newick.from_string("(((C:1,D:0.5):3.0,B:2):1.5,A:5.5);")
    assert_equal(expected, rerooted)


def test_newick__reroot_on_midpoint__nested_clades():
    source = Newick.from_string("((A:2,(B:2,C:3):4):1,(D:1,E:0.5):2);")
    rerooted = source.reroot_on_midpoint()
    expected = Newick.from_string("(((D:1,E:0.5):3.0,A:2):1.5,(B:2,C:3):2.5);")
    assert_equal(expected, rerooted)


def test_newick__reroot_on_midpoint__reroot_on_internal_node():
    source = Newick.from_string("((A:5.0,B:1.0)C:2.0,D:3.0);")
    rerooted = source.reroot_on_midpoint()
    expected = Newick.from_string("(A:5.0,B:1.0,D:5.0)C;")
    assert_equal(expected, rerooted)


def test_newick__reroot_on_midpoint__invalid_branch_lengths():
    def _test_invalid_branch_lengths(newick):
        source = Newick.from_string(newick)
        assert_raises(GraphError, source.reroot_on_midpoint)

    yield _test_invalid_branch_lengths, "(A,B);"  # No branch lengths
    yield _test_invalid_branch_lengths, "(A:7,B);"  # Length missing for leaf node
    yield _test_invalid_branch_lengths, "(A:7,(B:3));"  # Length missing for internal node
    yield _test_invalid_branch_lengths, "(A:7,(B:3):-1);"  # Negative branch length
    yield _test_invalid_branch_lengths, "(A:7,B:-1);"  # Negative leaf length


###############################################################################
###############################################################################
# add_support

def test_newick__add_support__no_trees():
    main_tree = Newick.from_string("(((A,B),C),D);")
    expected = Newick.from_string("(((A,B)0,C)0,D);")
    result = main_tree.add_support([])
    assert_equal(expected, result)


def test_newick__add_support__single_identical_tree():
    main_tree = Newick.from_string("(((A,B),C),D);")
    bootstraps = [Newick.from_string("(((A,B),C),D);")]
    expected = Newick.from_string("(((A,B)1,C)1,D);")
    result = main_tree.add_support(bootstraps)
    assert_equal(expected, result)


def test_newick__add_support__single_identical_tree__different_rooting():
    main_tree = Newick.from_string("(((A,B),C),D);")
    bootstraps = [Newick.from_string("(((C,D),B),A);")]
    expected = Newick.from_string("(((A,B)1,C)1,D);")
    result = main_tree.add_support(bootstraps)
    assert_equal(expected, result)


def test_newick__add_support__multiple_trees__different_topologies():
    main_tree = Newick.from_string("(((A,B),C),D);")
    bootstraps = [Newick.from_string("(((C,B),D),A);"),
                  Newick.from_string("(((A,D),B),C);")]
    expected = Newick.from_string("(((A,B)0,C)2,D);")
    result = main_tree.add_support(bootstraps)
    assert_equal(expected, result)


def test_newick__add_support__multiple_trees__partially_different_topologies():
    main_tree = Newick.from_string("(((A,B),C),D);")
    bootstraps = [Newick.from_string("(((C,D),A),B);"),
                  Newick.from_string("(((A,D),B),C);")]
    expected = Newick.from_string("(((A,B)1,C)2,D);")
    result = main_tree.add_support(bootstraps)
    assert_equal(expected, result)


def test_newick__add_support__multiple_trees__two_cladees():
    main_tree = Newick.from_string("((A,B),(C,(D,E)));")
    bootstraps = [Newick.from_string("((((C,E),D),A),B);"),
                  Newick.from_string("(((A,(C,D)),B),E);")]
    expected = Newick.from_string("((A,B)1,(C,(D,E)0)1);")
    result = main_tree.add_support(bootstraps)
    assert_equal(expected, result)


def test_newick__add_support__differing_leaf_names():
    main_tree = Newick.from_string("(((A,B),C),D);")
    bootstraps = [Newick.from_string("(((C,E),B),A);")]
    assert_raises(NewickError, main_tree.add_support, bootstraps)


def test_newick__add_support__formatting():
    def _do_test_formatting(fmt, expected):
        main_tree = Newick.from_string("(((A,B),C),D);")
        bootstraps = [Newick.from_string("(((C,D),A),B);"),
                      Newick.from_string("(((C,B),A),D);"),
                      Newick.from_string("(((A,D),B),C);")]
        expected = Newick.from_string(expected)
        result = main_tree.add_support(bootstraps, fmt)
        assert_equal(expected, result)

    yield _do_test_formatting, "{Support}",        "(((A,B)1,C)3,D);"
    yield _do_test_formatting, "{Percentage:.0f}", "(((A,B)33,C)100,D);"
    yield _do_test_formatting, "{Fraction:.2f}",   "(((A,B)0.33,C)1.00,D);"


def test_newick__add_support__unique_names_required():
    main_tree = Newick.from_string("(((A,B),C),A);")
    bootstraps = [Newick.from_string("(((A,B),C),A);")]
    assert_raises(NewickError, main_tree.add_support, bootstraps)


###############################################################################
###############################################################################
# from_string

def test_newick__parse__minimal_newick__name_only():
    top_node = Newick(name="A")
    assert_equal(Newick.from_string("A;"), top_node)


def test_newick__parse__single_taxa():
    child_node = Newick(name="Ab")
    top_node = Newick(children=[child_node])
    assert_equal(Newick.from_string("(Ab);"), top_node)


def test_newick__parse__two_taxa():
    child_node_1 = Newick(name="A")
    child_node_2 = Newick(name="Bc")
    top_node = Newick(children=[child_node_1, child_node_2])
    assert_equal(Newick.from_string("(A,Bc);"), top_node)


def test_newick__parse__three_taxa():
    child_node_1 = Newick(name="A")
    child_node_2 = Newick(name="Bc")
    child_node_3 = Newick(name="DeF")
    top_node = Newick(children=[child_node_1, child_node_2, child_node_3])
    assert_equal(Newick.from_string("(A,Bc,DeF);"), top_node)


def test_newick__parse__ignore_whitespace():
    assert_equal(Newick.from_string("(A,B);"), Newick.from_string("(A, B);"))


def test_newick__parse__missing_semicolon():
    assert_raises(NewickParseError, Newick.from_string, "()")


def test_newick__parse__subnode__single_taxa():
    child_node_1 = Newick(name="A")
    child_node_2a = Newick(name="B")
    child_node_2 = Newick(children=[child_node_2a])
    top_node = Newick(children=[child_node_1, child_node_2])
    assert_equal(Newick.from_string("(A,(B));"), top_node)


def test_newick__parse__subnode__two_taxa():
    child_node_1 = Newick(name="A")
    child_node_2a = Newick(name="B")
    child_node_2b = Newick(name="C")
    child_node_2 = Newick(children=[child_node_2a, child_node_2b])
    top_node = Newick(children=[child_node_1, child_node_2])
    assert_equal(Newick.from_string("(A,(B,C));"), top_node)


###########################################################################
###########################################################################
# cmp - white-box, just make sure all properties are compared

def test_newick__cmp__identical():
    node_a = Newick(name="A", length=13, children=[Newick(name="B")])
    node_b = Newick(name="A", length=13, children=[Newick(name="B")])
    assert_equal(node_a, node_b)


def test_newick__cmp__identical_for_empty_string_length():
    node_a = Newick(name="A", length="", children=[Newick(name="B")])
    node_b = Newick(name="A", length=None, children=[Newick(name="B")])
    assert_equal(node_a, node_b)


def test_newick__cmp__identical_for_empty_string_name():
    node_a = Newick(name="", length=13, children=[Newick(name="B")])
    node_b = Newick(name=None, length=13, children=[Newick(name="B")])
    assert_equal(node_a, node_b)


def test_newick__cmp__not_identical():
    def _not_identical(node_b):
        node_a = Newick(name="A", length=13, children=[Newick(name="B")])
        assert_not_equal(node_a, node_b)
    yield _not_identical, Newick(name="B", length=13, children=[Newick(name="B")])
    yield _not_identical, Newick(name="A", length=14, children=[Newick(name="B")])
    yield _not_identical, Newick(name="A", length=13, children=[])
    yield _not_identical, Newick(name="A", length=13, children=[Newick(name="C")])
    yield _not_identical, Newick(name="B", length=14, children=[Newick(name="C")])


###############################################################################
###############################################################################
# hash - white-box, just make sure all properties are used

def test_newick__hash__identical():
    node_a = Newick(name="A", length=13, children=[Newick(name="B")])
    node_b = Newick(name="A", length=13, children=[Newick(name="B")])
    assert_equal(hash(node_a), hash(node_b))


def test_newick__hash__not_identical():
    def _not_identical(node_b):
        node_a = Newick(name="A", length=13, children=[Newick(name="B")])
        assert_not_equal(hash(node_a), hash(node_b))
    yield _not_identical, Newick(name="B", length=13, children=[Newick(name="B")])
    yield _not_identical, Newick(name="A", length=14, children=[Newick(name="B")])
    yield _not_identical, Newick(name="A", length=13, children=[])
    yield _not_identical, Newick(name="A", length=13, children=[Newick(name="C")])
    yield _not_identical, Newick(name="B", length=14, children=[Newick(name="C")])


def test_newick__hash__hashable():
    key_a = Newick(name="A", length=13.7, children=[Newick(name="F")])
    key_b = Newick(name="A", length=13.7, children=[Newick(name="F")])
    assert key_b in {key_a: True}


###############################################################################
###############################################################################
# Malformed newick strings

def test_newick__malformed__unbalanced_parantheses():
    assert_raises(NewickParseError, Newick.from_string, "(A,(B,C);")


def test_newick__malformed__mismatched_parantheses():
    assert_raises(NewickParseError, Newick.from_string, "(A,(B,C();")


def test_newick__malformed__missing_parantheses():
    assert_raises(NewickParseError, Newick.from_string, "(A,(B,C))")


def test_newick__malformed__missing_length():
    assert_raises(NewickParseError, Newick.from_string, "(A:,(B,C));")
    assert_raises(NewickParseError, Newick.from_string, "(A,(B:,C));")
    assert_raises(NewickParseError, Newick.from_string, "(A,(B,C:));")
    assert_raises(NewickParseError, Newick.from_string, "(A,(B,C):);")
    assert_raises(NewickParseError, Newick.from_string, "(A,(B,C)):;")


def test_newick__malformed__multiple_lengths():
    assert_raises(NewickParseError, Newick.from_string, "(A:1:2);")


###############################################################################
###############################################################################
# Implicit leafs are not supported (due to problems with ambiguiety)

def test_newick__parse__first_taxa_unnamed():
    assert_raises(NewickError, Newick.from_string, "(,A);")


def test_newick__parse__second_taxa_unnamed():
    assert_raises(NewickError, Newick.from_string, "(A,);")


def test_newick__parse__two_taxa_unnamed():
    assert_raises(NewickError, Newick.from_string, "(,);")


def test_newick__parse__three_taxa_unnamed():
    assert_raises(NewickError, Newick.from_string, "(,,);")


###############################################################################
###############################################################################
# Empty non-leaf nodes are not allowed, as their interpretation is unclear

def test_newick__parse__minimal_newick__implicit_nodes():
    assert_raises(NewickParseError, Newick.from_string, "();")


def test_newick__parse__subnode__empty():
    assert_raises(NewickParseError, Newick.from_string, "(A,());")


###############################################################################
###############################################################################
# The following tests are derived from the wikipedia description of the
# newick format: http://en.wikipedia.org/wiki/Newick_format#Examples

def test_newick__wikipedia_example_1():
    # no nodes are named, this format is not supported here!
    assert_raises(NewickError, Newick.from_string, "(,,(,));")


def test_newick__wikipedia_example_2():
    # leaf nodes are named
    taxa_d = Newick(name="D")
    taxa_c = Newick(name="C")
    taxa_sub = Newick(children=[taxa_c, taxa_d])
    taxa_b = Newick(name="B")
    taxa_a = Newick(name="A")
    top_node = Newick(children=[taxa_a, taxa_b, taxa_sub])
    assert_equal(Newick.from_string("(A,B,(C,D));"), top_node)


def test_newick__wikipedia_example_3():
    # all nodes are named
    taxa_d = Newick(name="D")
    taxa_c = Newick(name="C")
    taxa_sub = Newick(children=[taxa_c, taxa_d], name="E")
    taxa_b = Newick(name="B")
    taxa_a = Newick(name="A")
    top_node = Newick(children=[taxa_a, taxa_b, taxa_sub], name="F")
    assert_equal(Newick.from_string("(A,B,(C,D)E)F;"), top_node)


def test_newick__wikipedia_example_4():
    # all but root node have a distance to parent
    taxa_d = Newick(length="0.4")
    taxa_c = Newick(length="0.3")
    taxa_sub = Newick(children=[taxa_c, taxa_d], length="0.5")
    taxa_b = Newick(length="0.2")
    taxa_a = Newick(length="0.1")
    top_node = Newick(children=[taxa_a, taxa_b, taxa_sub])
    assert_equal(Newick.from_string("(:0.1,:0.2,(:0.3,:0.4):0.5);"), top_node)


def test_newick__wikipedia_example_5():
    # all have a distance to parent
    taxa_d = Newick(length="0.4")
    taxa_c = Newick(length="0.3")
    taxa_sub = Newick(children=[taxa_c, taxa_d], length="0.5")
    taxa_b = Newick(length="0.2")
    taxa_a = Newick(length="0.1")
    top_node = Newick(children=[taxa_a, taxa_b, taxa_sub], length="0.0")
    assert_equal(Newick.from_string("(:0.1,:0.2,(:0.3,:0.4):0.5):0.0;"), top_node)


def test_newick__wikipedia_example_6():
    # distances and leaf names (popular)
    taxa_d = Newick(length="0.4", name="D")
    taxa_c = Newick(length="0.3", name="C")
    taxa_sub = Newick(children=[taxa_c, taxa_d], length="0.5")
    taxa_b = Newick(length="0.2", name="B")
    taxa_a = Newick(length="0.1", name="A")
    top_node = Newick(children=[taxa_a, taxa_b, taxa_sub])
    assert_equal(Newick.from_string("(A:0.1,B:0.2,(C:0.3,D:0.4):0.5);"), top_node)


def test_newick__wikipedia_example_7():
    # distances and all names
    taxa_d = Newick(length="0.4", name="D")
    taxa_c = Newick(length="0.3", name="C")
    taxa_sub = Newick(children=[taxa_c, taxa_d], length="0.5", name="E")
    taxa_b = Newick(length="0.2", name="B")
    taxa_a = Newick(length="0.1", name="A")
    top_node = Newick(children=[taxa_a, taxa_b, taxa_sub], name="F")
    assert_equal(Newick.from_string("(A:0.1,B:0.2,(C:0.3,D:0.4)E:0.5)F;"), top_node)


def test_newick__wikipedia_example_8():
    # a tree rooted on a leaf node (rare)
    taxa_b = Newick(length="0.2", name="B")
    taxa_c = Newick(length="0.3", name="C")
    taxa_d = Newick(length="0.4", name="D")
    node_e = Newick(length="0.5", name="E", children=[taxa_c, taxa_d])
    node_f = Newick(length="0.1", name="F", children=[taxa_b, node_e])
    node_a = Newick(name="A", children=[node_f])
    assert_equal(Newick.from_string("((B:0.2,(C:0.3,D:0.4)E:0.5)F:0.1)A;"), node_a)


###############################################################################
###############################################################################
# str / repr

def test_newick__str__non_string_name():
    node = Newick(children=[Newick(name=17, length="1.3")])
    assert_equal(str(node), "(17:1.3);")


def test_newick__str__non_string_length():
    node = Newick(children=[Newick(name="Foo", length=1.3)])
    assert_equal(str(node), "(Foo:1.3);")


def test_newick__str__repr_equal_to_str():
    node_a = Newick(name="A", length="123")
    node_b = Newick(name="B", length="4567")
    top_node = Newick(children=[node_a, node_b])
    assert_equal(str(top_node), "(A:123,B:4567);")


def test_newick__str__single_leaf_should_not_be_followed_by_comma():
    node = Newick(name="A")
    top_node = Newick(children=[node])
    assert_equal(str(top_node), "(A);")


def test_newick__wikipedia_examples__str_equality():
    def _newick_str_input_equals_output(nwk_str):
        nodes = Newick.from_string(nwk_str)
        result = str(nodes)
        assert_equal(result, nwk_str)

    # 2. leaf nodes are named
    yield _newick_str_input_equals_output, "(A,B,(C,D));"
    # 3. all nodes are named
    yield _newick_str_input_equals_output, "(A,B,(C,D)E)F;"
    # 4. all but root node have a distance to parent
    yield _newick_str_input_equals_output, "(:0.1,:0.2,(:0.3,:0.4):0.5);"
    # 5. all have a distance to parent
    yield _newick_str_input_equals_output, "(:0.1,:0.2,(:0.3,:0.4):0.5):0.0;"
    # 6. distances and leaf names (popular)
    yield _newick_str_input_equals_output, "(A:0.1,B:0.2,(C:0.3,D:0.4):0.5);"
    # 7. distances and all names
    yield _newick_str_input_equals_output, "(A:0.1,B:0.2,(C:0.3,D:0.4)E:0.5)F;"


###############################################################################
###############################################################################
# Immutability

def test_newick__properties_are_immutable():
    def _test_newick__properties_are_immutable(name, value):
        node = Newick(name="A", length=3, children=[Newick(name="B")])
        assert_raises(NotImplementedError, setattr, node, name, value)
    yield _test_newick__properties_are_immutable, "name", "foo"
    yield _test_newick__properties_are_immutable, "length", "13"
    yield _test_newick__properties_are_immutable, "children", []
    yield _test_newick__properties_are_immutable, "foobar", True


def test_newick__properties_cannot_be_deleted():
    def _test_newick__properties_cannot_be_deleted(name):
        node = Newick(name="A", length=3, children=[Newick(name="B")])
        assert_raises(NotImplementedError, delattr, node, name)
    yield _test_newick__properties_cannot_be_deleted, "name"
    yield _test_newick__properties_cannot_be_deleted, "length"
    yield _test_newick__properties_cannot_be_deleted, "children"
    yield _test_newick__properties_cannot_be_deleted, "foobar"



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

from paleomix.common.utilities import safe_coerce_to_tuple
from paleomix.nodes.picard import MergeSamFilesNode
from paleomix.tools.bam_pipeline.nodes import \
    index_and_validate_bam
from paleomix.nodes.validation import \
    DetectInputDuplicationNode

import paleomix.ui as ui
import paleomix.nodes.gatk as gatk


class Prefix(object):
    def __init__(self, config, prefix, samples, features, target):
        self.name = prefix["Name"]
        self.label = prefix.get("Label") or self.name
        self.roi = prefix.get("RegionsOfInterest", {})

        self.samples = safe_coerce_to_tuple(samples)
        self.folder = config.destination
        self.target = target

        files_and_nodes = {}
        for sample in self.samples:
            files_and_nodes.update(sample.bams.iteritems())

        self.datadup_check = self._build_dataduplication_node(
            prefix, files_and_nodes)

        build_raw_bam = features["RawBAM"]
        build_realigned_bam = features["RealignedBAM"]
        if build_realigned_bam and prefix['IndexFormat'] == '.csi':
            if prefix['Path'] not in _CSI_WARNINGS:
                ui.print_err("\nWARNING: Realigned BAMs enabled for reference "
                             "genome %r, but the file contains sequences too "
                             "large for GATK, which does not support .csi "
                             "index files. Raw BAMs will be built instead of "
                             "realigned BAMs, for this reference sequence."
                             % (prefix['Path']))

                # TODO: Add reference to FAQ when written.

            _CSI_WARNINGS.add(prefix['Path'])
            build_realigned_bam = False
            build_raw_bam = True

        self.bams = {}
        if build_raw_bam:
            self.bams.update(self._build_raw_bam(
                config, prefix, files_and_nodes))

        if build_realigned_bam:
            self.bams.update(self._build_realigned_bam(
                config, prefix, files_and_nodes))

        if not self.bams:
            for sample in self.samples:
                self.bams.update(sample.bams)

        nodes = [self.datadup_check]
        for sample in self.samples:
            nodes.extend(sample.nodes)
        self.nodes = tuple(nodes)

    def _build_raw_bam(self, config, prefix, files_and_bams):
        output_filename = os.path.join(
            self.folder, "%s.%s.bam" % (self.target, prefix["Name"]))
        validated_filename = os.path.join(
            self.folder, self.target, prefix["Name"] + ".validated")

        node = MergeSamFilesNode(config=config,
                                 input_bams=files_and_bams.keys(),
                                 output_bam=output_filename,
                                 dependencies=self.datadup_check)
        validated_node = index_and_validate_bam(config=config,
                                                prefix=prefix,
                                                node=node,
                                                log_file=validated_filename)

        return {output_filename: validated_node}

    def _build_realigned_bam(self, config, prefix, bams):
        output_filename = os.path.join(
            self.folder, "%s.%s.realigned.bam" % (self.target, prefix["Name"]))
        intervals_filename = os.path.join(
            self.folder, self.target, prefix["Name"] + ".intervals")
        validated_filename = os.path.join(
            self.folder, self.target, prefix["Name"] + ".realigned.validated")

        trainer = gatk.GATKIndelTrainerNode(config=config,
                                            reference=prefix["Reference"],
                                            infiles=bams.keys(),
                                            outfile=intervals_filename,
                                            threads=config.gatk_max_threads,
                                            dependencies=self.datadup_check)

        aligner = gatk.GATKIndelRealignerNode(config=config,
                                              reference=prefix["Reference"],
                                              infiles=bams.keys(),
                                              intervals=intervals_filename,
                                              outfile=output_filename,
                                              dependencies=trainer)

        validated_node = index_and_validate_bam(config=config,
                                                prefix=prefix,
                                                node=aligner,
                                                log_file=validated_filename)

        return {output_filename: validated_node}

    def _build_dataduplication_node(self, prefix, files_and_nodes):
        filename = prefix["Name"] + ".duplications_checked"
        destination = os.path.join(self.folder, self.target, filename)
        dependencies = files_and_nodes.values()

        return DetectInputDuplicationNode(input_files=files_and_nodes.keys(),
                                          output_file=destination,
                                          dependencies=dependencies)


# Contains the paths of sequences for which warnings about GATK has been given,
# if the 'RealignedBAM' feature was enabled for files that require a CSI index.
_CSI_WARNINGS = set()



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# pylint: disable=missing-docstring
#
import os
import signal
import weakref

from flexmock import flexmock

import nose
from nose.tools import \
    assert_in, \
    assert_equal, \
    assert_raises

from paleomix.common.testing import \
    with_temp_folder, \
    Monkeypatch, \
    get_file_contents, \
    set_file_contents

import paleomix.atomiccmd.command
import paleomix.common.fileutils as fileutils

from paleomix.common.versions import RequirementObj
from paleomix.atomiccmd.command import AtomicCmd, CmdError


def test_file(*args):
    test_root = os.path.dirname(os.path.dirname(__file__))

    return os.path.join(test_root, "data", *args)


###############################################################################
###############################################################################
# Constructor: Command

def test_atomiccmd__command_str():
    cmd = AtomicCmd("ls")
    assert_equal(cmd.executables, frozenset(["ls"]))


@nose.tools.raises(ValueError)
def test_atomiccmd__executables_empty_str():
    AtomicCmd("")


def test_atomiccmd__command_tuple():
    cmd = AtomicCmd(("cd", "."))
    assert_equal(cmd.executables, frozenset(["cd"]))


@nose.tools.raises(ValueError)
def test_atomiccmd__executables_empty_tuple():
    AtomicCmd(())


@nose.tools.raises(ValueError)
def test_atomiccmd__executables_empty_str_in_tuple():
    AtomicCmd((""))


###############################################################################
###############################################################################
# Constructor: set_cwd

def test_atomiccmd__set_cwd():
    @with_temp_folder
    def _do_test_atomiccmd__set_cwd(temp_folder, set_cwd):
        cwd = os.getcwd()
        cmd = AtomicCmd(("bash", "-c", "echo -n ${PWD}"),
                        TEMP_OUT_STDOUT="result.txt",
                        set_cwd=set_cwd)
        cmd.run(temp_folder)
        assert_equal(cmd.join(), [0])
        assert_equal(cwd, os.getcwd())

        expected = temp_folder if set_cwd else cwd
        result = get_file_contents(os.path.join(temp_folder, "result.txt"))
        assert os.path.samefile(
            expected, result), "%r != %r" % (expected, result)

    yield _do_test_atomiccmd__set_cwd, False
    yield _do_test_atomiccmd__set_cwd, True


# Full path when set_cwd is False, rel. path when True
def test_atomiccmd__set_cwd__temp_in_out():
    @with_temp_folder
    def _test_atomiccmd__paths_temp_in(temp_folder, set_cwd, kwargs):
        cmd = AtomicCmd(("echo", "-n", "%%(%s)s" % tuple(kwargs.keys())),
                        TEMP_OUT_STDOUT="result.txt",
                        set_cwd=set_cwd,
                        **kwargs)
        cmd.run(temp_folder)
        assert_equal(cmd.join(), [0])

        expected = os.path.join("" if set_cwd else temp_folder, "test_file")
        result = get_file_contents(os.path.join(temp_folder, "result.txt"))
        assert_equal(os.path.abspath(expected), os.path.abspath(result))

    yield _test_atomiccmd__paths_temp_in, True, {"TEMP_IN_FOO": "test_file"}
    yield _test_atomiccmd__paths_temp_in, False, {"TEMP_IN_FOO": "test_file"}
    yield _test_atomiccmd__paths_temp_in, True, {"TEMP_OUT_FOO": "test_file"}
    yield _test_atomiccmd__paths_temp_in, False, {"TEMP_OUT_FOO": "test_file"}


###############################################################################
###############################################################################
# Constructor: Paths / pipes

# Check that specified paths/etc. are available via getters
def test_atomiccmd__paths():
    cmd = AtomicCmd("ls",
                    IN_AAA="/a/b/c",
                    IN_AAB="/x/y/z",
                    TEMP_IN_ABB="tmp_in",
                    OUT_AAA="/out/foo",
                    OUT_BBC="foo/bar",
                    TEMP_OUT_A="xyb",
                    EXEC_OTHER="true",
                    AUX_WAT="wat/wat",
                    CHECK_FUNC=bool,
                    OUT_STDERR="/var/log/pipe.stderr",
                    TEMP_OUT_STDOUT="pipe.stdout")

    assert_equal(cmd.executables, frozenset(["ls", "true"]))
    assert_equal(cmd.requirements, frozenset([bool]))
    assert_equal(cmd.input_files, frozenset(["/a/b/c", "/x/y/z"]))
    assert_equal(cmd.output_files, frozenset(
        ["/out/foo", "foo/bar", "/var/log/pipe.stderr"]))
    assert_equal(cmd.auxiliary_files, frozenset(["wat/wat"]))
    assert_equal(cmd.expected_temp_files, frozenset(
        ["foo", "bar", "pipe.stderr"]))
    assert_in("xyb", cmd.optional_temp_files)
    assert_in("pipe.stdout", cmd.optional_temp_files)


def test_atomiccmd__paths_optional():
    cmd = AtomicCmd(["ls"],
                    IN_OPTIONAL=None,
                    OUT_OPTIONAL=None)
    assert_equal(cmd.input_files, frozenset())
    assert_equal(cmd.output_files, frozenset())


@with_temp_folder
def test_atomiccmd__pipes_stdin(temp_folder):
    fname = test_file("fasta_file.fasta")
    cmd = AtomicCmd("cat",
                    IN_STDIN=fname,
                    OUT_STDOUT="result.txt")
    assert_equal(cmd.input_files, frozenset([fname]))
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])
    result = get_file_contents(os.path.join(temp_folder, "result.txt"))
    assert_equal(
        result, ">This_is_FASTA!\nACGTN\n>This_is_ALSO_FASTA!\nCGTNA\n")


@with_temp_folder
def test_atomiccmd__pipes_stdin__temp_file(temp_folder):
    cmd = AtomicCmd("cat",
                    TEMP_IN_STDIN="infile.fasta",
                    OUT_STDOUT="result.txt")
    assert_equal(cmd.input_files, frozenset())
    set_file_contents(os.path.join(temp_folder, "infile.fasta"), "a\nbc\nd")
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])
    result = get_file_contents(os.path.join(temp_folder, "result.txt"))
    assert_equal(result, "a\nbc\nd")


@with_temp_folder
def test_atomiccmd__pipes_stdin__dev_null_implicit_1(temp_folder):
    # STDIN should be implicitly set to /dev/null; deadlocks if not
    cmd = AtomicCmd("cat")
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])


@with_temp_folder
def test_atomiccmd__pipes_stdin__dev_null_implicit_2(temp_folder):
    # STDIN should be implicitly set to /dev/null; deadlocks if not
    cmd = AtomicCmd("cat", IN_STDIN=None)
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])


@with_temp_folder
def test_atomiccmd__pipes_stdin__dev_null_explicit(temp_folder):
    # STDIN should be set to /dev/null; deadlocks if not
    cmd = AtomicCmd("cat", IN_STDIN=AtomicCmd.DEVNULL)
    cmd.run(temp_folder, wrap_errors=False)
    assert_equal(cmd.join(), [0])


# Test possible combinations of explicit / implicit saving of stdout/err
def test_atomiccmd__pipes_out():
    @with_temp_folder
    def _test_atomiccmd__pipes_out(temp_folder, stdout, stderr, kwargs):
        call = ("bash", "-c",
                "echo -n 'STDERR!' > /dev/stderr; echo -n 'STDOUT!';")

        cmd = AtomicCmd(call, **kwargs)
        cmd.run(temp_folder)
        assert_equal(cmd.join(), [0])

        expected_files = []
        for (tmpl, text) in ((stdout, "STDOUT!"), (stderr, "STDERR!")):
            if tmpl is not None:
                fname = tmpl.format(id(cmd))
                result = get_file_contents(os.path.join(temp_folder, fname))
                assert_equal(result, text)
                expected_files.append(fname)

        assert_equal(set(os.listdir(temp_folder)), set(expected_files))

    yield _test_atomiccmd__pipes_out, \
        "pipe_bash_{0}.stdout", "pipe_bash_{0}.stderr", {}
    yield _test_atomiccmd__pipes_out, \
        "pipe_bash_{0}.stdout", "stderr.txt", {"OUT_STDERR": "stderr.txt"}
    yield _test_atomiccmd__pipes_out, \
        "stdout.txt", "pipe_bash_{0}.stderr", {"OUT_STDOUT": "stdout.txt"}
    yield _test_atomiccmd__pipes_out, \
        "stdout.txt", "stderr.txt", {"OUT_STDOUT": "stdout.txt",
                                     "OUT_STDERR": "stderr.txt"}

    yield _test_atomiccmd__pipes_out, \
        None, None, {"OUT_STDOUT": AtomicCmd.DEVNULL,
                     "OUT_STDERR": AtomicCmd.DEVNULL}
    yield _test_atomiccmd__pipes_out, \
        None, "pipe_bash_{0}.stderr", {"OUT_STDOUT": AtomicCmd.DEVNULL}
    yield _test_atomiccmd__pipes_out, \
        "pipe_bash_{0}.stdout", None, {"OUT_STDERR": AtomicCmd.DEVNULL}


def test_atomiccmd__paths__malformed_keys():
    def _do_test_atomiccmd__paths__malformed(kwargs):
        assert_raises(ValueError, AtomicCmd, "true", **kwargs)

    # Missing key-name #1
    yield _do_test_atomiccmd__paths__malformed, {"IN": "/var/foo"}
    # Missing key-name #2
    yield _do_test_atomiccmd__paths__malformed, {"IN_": "/var/foo"}
    # Missing key-name #3
    yield _do_test_atomiccmd__paths__malformed, {"TEMP_OUT": "/var/foo"}
    # Missing key-name #4
    yield _do_test_atomiccmd__paths__malformed, {"TEMP_OUT_": "/var/foo"}
    # Invalid key-type #1
    yield _do_test_atomiccmd__paths__malformed, {"TEMP_OUX_FOO": "foo"}
    # Invalid key-type #2
    yield _do_test_atomiccmd__paths__malformed, {"INS_BAR": "foo"}


def test_atomiccmd__paths__invalid_values():
    def _do_test_atomiccmd__paths__invalid_values(kwargs):
        assert_raises(TypeError, AtomicCmd, "true", **kwargs)

    # Long name is used to create more informative error messages
    test_function = _do_test_atomiccmd__paths__invalid_values

    yield test_function, {"IN_FILE": 1}
    yield test_function, {"TEMP_IN_FILE": set()}
    yield test_function, {"OUT_FILE": [1, 2, 3]}
    yield test_function, {"TEMP_OUT_FILE": 1.0}

    yield test_function, {"IN_STDIN": {}}
    yield test_function, {"TEMP_IN_STDIN": frozenset()}
    yield test_function, {"OUT_STDOUT": 1.7}
    yield test_function, {"TEMP_OUT_STDOUT": ()}
    yield test_function, {"OUT_STDERR": xrange(3)}
    yield test_function, {"TEMP_OUT_STDERR": -1}


# Subpaths are not allowed for temp IN/OUT files, neither relative nor
# asbsolute
def test_atomiccmd__paths__invalid_temp_paths():
    def _do_test_atomiccmd__paths__invalid_temp_paths(kwargs):
        assert_raises(ValueError, AtomicCmd, "true", **kwargs)

    # Long name is used to create more informative error messages
    test_function = _do_test_atomiccmd__paths__invalid_temp_paths

    # No relative paths
    yield test_function, {"TEMP_IN_FOO": "sub/infile"}
    yield test_function, {"TEMP_IN_STDIN": "sub/stdin"}
    yield test_function, {"TEMP_OUT_FOO": "sub/outfile"}
    yield test_function, {"TEMP_OUT_STDOUT": "sub/stdout"}
    yield test_function, {"TEMP_OUT_STDERR": "sub/stderr"}

    # No absolute paths
    yield test_function, {"TEMP_IN_FOO": "/tmp/sub/infile"}
    yield test_function, {"TEMP_IN_STDIN": "/dev/sub/stdin"}
    yield test_function, {"TEMP_OUT_FOO": "/etc/sub/outfile"}
    yield test_function, {"TEMP_OUT_STDOUT": "/var/sub/stdout"}
    yield test_function, {"TEMP_OUT_STDERR": "/home/sub/stderr"}


# All OUT_ files must be unique, including all TEMP_OUT_
def test_atomiccmd__paths__overlapping_output():
    def _do_test_atomiccmd__paths__overlapping_output(key_1, file_1,
                                                      key_2, file_2):
        assert_raises(ValueError, AtomicCmd, ("ls",), **{key_1: file_1,
                                                         key_2: file_2})

    test = _do_test_atomiccmd__paths__overlapping_output

    yield test, "OUT_FILE_1", "/foo/bar/outfile", "OUT_FILE_2", "/var/outfile"
    yield test, "TEMP_OUT_FILE_1", "outfile", "OUT_FILE_1", "/var/outfile"
    yield test, "OUT_FILE_1", "/foo/bar/outfile", "TEMP_OUT_FILE_1", "outfile"
    yield test, "TEMP_OUT_FILE_1", "outfile", "TEMP_OUT_FILE_2", "outfile"

    yield test, "OUT_FILE_1", "/foo/bar/outfile", "OUT_STDOUT", "/var/outfile"
    yield test, "TEMP_OUT_FILE_1", "outfile", "OUT_STDOUT", "/var/outfile"
    yield test, "OUT_FILE_1", "/foo/bar/outfile", "TEMP_OUT_STDOUT", "outfile"
    yield test, "TEMP_OUT_FILE_1", "outfile", "TEMP_OUT_STDOUT", "outfile"

    yield test, "OUT_FILE_1", "/foo/bar/outfile", "OUT_STDERR", "/var/outfile"
    yield test, "TEMP_OUT_FILE_1", "outfile", "OUT_STDERR", "/var/outfile"
    yield test, "OUT_FILE_1", "/foo/bar/outfile", "TEMP_OUT_STDERR", "outfile"
    yield test, "TEMP_OUT_FILE_1", "outfile", "TEMP_OUT_STDERR", "outfile"


# A pipe can be w/wo TEMP_, but not both
def test_atomiccmd__pipes__duplicates():
    def _do_test_atomiccmd__pipes__duplicates(key):
        kwargs = {"TEMP_" + key: "temp_file",
                  key: "file"}
        assert_raises(CmdError, AtomicCmd, ["ls"], **kwargs)

    yield _do_test_atomiccmd__pipes__duplicates, "IN_STDIN"
    yield _do_test_atomiccmd__pipes__duplicates, "OUT_STDOUT"
    yield _do_test_atomiccmd__pipes__duplicates, "OUT_STDERR"


###############################################################################
###############################################################################
# CHECK_ / EXEC_

# RequirementObjs are the standard way to do tests
def test_atomicmcd__exec__reqobj():
    reqobj = RequirementObj(call=("echo", "version"),
                            search="version",
                            checks=str)
    cmd = AtomicCmd("true",
                    CHECK_VERSION=reqobj)
    assert_equal(cmd.requirements, frozenset([reqobj]))


# CHECK_ is expected to be a callable
@nose.tools.raises(TypeError)
def test_atomiccmd__checks__non_callable():
    AtomicCmd("ls", CHECK_FOO="ls")


# EXEC_ is expected to be a string
def test_atomiccmd__exec__invalid():
    @nose.tools.raises(TypeError)
    def _test_atomiccmd__exec__invalid(obj):
        AtomicCmd("true", EXEC_FOO=obj)

    yield _test_atomiccmd__exec__invalid, str
    yield _test_atomiccmd__exec__invalid, {}
    yield _test_atomiccmd__exec__invalid, 1


###############################################################################
###############################################################################
# AUX

def test_atomiccmd__aux__invalid():
    @nose.tools.raises(TypeError)
    def _test_atomiccmd__exec__invalid(obj):
        AtomicCmd("true", AUX_FOO=obj)

    yield _test_atomiccmd__exec__invalid, str
    yield _test_atomiccmd__exec__invalid, {}
    yield _test_atomiccmd__exec__invalid, 1


###############################################################################
###############################################################################
# Path components

@with_temp_folder
def test_atomiccmd__paths_non_str(temp_folder):
    cmd = AtomicCmd(("touch", 1234),
                    OUT_FOO="1234",
                    set_cwd=True)
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])
    assert os.path.exists(os.path.join(temp_folder, "1234"))


@nose.tools.raises(CmdError)
def test_atomiccmd__paths_missing():
    AtomicCmd(("touch", "%(IN_FOO)s"))


@nose.tools.raises(CmdError)
def test_atomiccmd__paths_invalid():
    AtomicCmd(("touch", "%(IN_FOO)"),
              IN_FOO="abc")


@with_temp_folder
def test_atomiccmd__paths__key(temp_folder):
    cmd = AtomicCmd(("echo", "-n", "%(TEMP_DIR)s"),
                    OUT_STDOUT=AtomicCmd.PIPE)
    cmd.run(temp_folder)
    path = cmd._proc.stdout.read()
    assert os.path.samefile(temp_folder, path), (temp_folder, path)
    assert_equal(cmd.join(), [0])


###############################################################################
###############################################################################
# Constructor: Piping commands

@with_temp_folder
def test_atomiccmd__piping(temp_folder):
    cmd_1 = AtomicCmd(["echo", "-n", "#@!$^"],
                      OUT_STDOUT=AtomicCmd.PIPE)
    assert_equal(cmd_1.output_files, frozenset())
    cmd_2 = AtomicCmd(["cat"],
                      IN_STDIN=cmd_1,
                      OUT_STDOUT="piped.txt")
    assert_equal(cmd_2.input_files, frozenset())
    cmd_1.run(temp_folder)
    cmd_2.run(temp_folder)
    assert_equal(cmd_1.join(), [0])
    assert_equal(cmd_2.join(), [0])
    result = get_file_contents(os.path.join(temp_folder, "piped.txt"))
    assert_equal(result, "#@!$^")


@with_temp_folder
def test_atomiccmd__piping_temp(temp_folder):
    cmd_1 = AtomicCmd(["echo", "-n", "#@!$^"],
                      TEMP_OUT_STDOUT=AtomicCmd.PIPE)
    assert_equal(cmd_1.output_files, frozenset())
    cmd_2 = AtomicCmd(["cat"],
                      TEMP_IN_STDIN=cmd_1,
                      OUT_STDOUT="piped.txt")
    assert_equal(cmd_2.input_files, frozenset())
    cmd_1.run(temp_folder)
    cmd_2.run(temp_folder)
    assert_equal(cmd_1.join(), [0])
    assert_equal(cmd_2.join(), [0])
    result = get_file_contents(os.path.join(temp_folder, "piped.txt"))
    assert_equal(result, "#@!$^")


# Only STDOUT takes AtomicCmd.PIPE
def test_atomiccmd__piping__wrong_pipe():
    def _test_atomiccmd__piping__wrong_pipe(key):
        assert_raises(TypeError, AtomicCmd, "ls", **{key: AtomicCmd.PIPE})

    yield _test_atomiccmd__piping__wrong_pipe, "IN_STDIN"
    yield _test_atomiccmd__piping__wrong_pipe, "TEMP_IN_STDIN"
    yield _test_atomiccmd__piping__wrong_pipe, "OUT_STDERR"
    yield _test_atomiccmd__piping__wrong_pipe, "TEMP_OUT_STDERR"


@with_temp_folder
def test_atomiccmd__piping_is_only_allowed_once(temp_folder):
    cmd_1 = AtomicCmd(["echo", "-n", "foo\nbar"],
                      OUT_STDOUT=AtomicCmd.PIPE)
    cmd_2a = AtomicCmd(["grep", "foo"],
                       IN_STDIN=cmd_1)
    cmd_2b = AtomicCmd(["grep", "bar"],
                       IN_STDIN=cmd_1)
    cmd_1.run(temp_folder)
    cmd_2a.run(temp_folder)
    assert_raises(CmdError, cmd_2b.run, temp_folder)
    assert_equal(cmd_1.join(), [0])
    assert_equal(cmd_2a.join(), [0])
    assert_equal(cmd_2b.join(), [None])


###############################################################################
###############################################################################
# run

@with_temp_folder
def test_atomiccmd__run__already_running(temp_files):
    cmd = AtomicCmd(("sleep", "10"))
    cmd.run(temp_files)
    assert_raises(CmdError, cmd.run, temp_files)
    cmd.terminate()
    cmd.join()


@with_temp_folder
def test_atomiccmd__run__exception_on_missing_command(temp_files):
    cmd = AtomicCmd(("xyzabcefgh", "10"))
    assert_raises(CmdError, cmd.run, temp_files)
    cmd.terminate()
    cmd.join()


@with_temp_folder
def test_atomiccmd__run__exception_on_missing_command__no_wrap(temp_files):
    cmd = AtomicCmd(("xyzabcefgh", "10"))
    assert_raises(OSError, cmd.run, temp_files, wrap_errors=False)
    cmd.terminate()
    cmd.join()


@with_temp_folder
def test_atomiccmd__run__invalid_temp(temp_files):
    cmd = AtomicCmd(("sleep", "10"))
    assert_raises(CmdError, cmd.run, os.path.join(temp_files, "foo"))
    cmd.terminate()
    cmd.join()


###############################################################################
###############################################################################
# Ready

@with_temp_folder
def test_atomiccmd__ready(temp_folder):
    cmd = AtomicCmd("ls")
    assert_equal(cmd.join(), [None])
    assert not cmd.ready()
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])
    assert cmd.ready()


###############################################################################
###############################################################################
# Join / wait

def test_atomiccmd__join_wait():
    @with_temp_folder
    def _do_test_atomiccmd__join_wait(temp_folder, func, call,
                                      before_run, after_run):
        cmd = AtomicCmd(call)
        assert_equal(func(cmd), before_run)
        cmd.run(temp_folder)
        assert_equal(func(cmd), after_run)

    yield _do_test_atomiccmd__join_wait, AtomicCmd.join, "true", [None], [0]
    yield _do_test_atomiccmd__join_wait, AtomicCmd.join, "false", [None], [1]

    yield _do_test_atomiccmd__join_wait, AtomicCmd.wait, "true", None, 0
    yield _do_test_atomiccmd__join_wait, AtomicCmd.wait, "false", None, 1


###############################################################################
###############################################################################
# Terminate

def test_atomiccmd__terminate():
    @with_temp_folder
    def _do_test_atomiccmd__terminate(temp_folder, raise_on_terminate):
        cmd = AtomicCmd(("sleep", "10"))
        cmd.run(temp_folder)

        killpg_was_called = []

        def _wrap_killpg(pid, sig):
            assert_equal(pid, cmd._proc.pid)
            assert_equal(sig, signal.SIGTERM)
            killpg_was_called.append(True)
            if raise_on_terminate:
                raise OSError("KABOOM!")

        with Monkeypatch("os.killpg", _wrap_killpg):
            cmd.terminate()
        cmd.terminate()
        assert_equal(cmd.join(), ["SIGTERM"])
        assert killpg_was_called
    yield _do_test_atomiccmd__terminate, False
    yield _do_test_atomiccmd__terminate, True


# Ensure that no OSException is raised, even if the command
# managed to finish before terminate was called
@with_temp_folder
def test_atomiccmd__terminate_race_condition(temp_folder):
    cmd = AtomicCmd("true")
    cmd.run(temp_folder)
    while cmd._proc.poll() is None:
        pass
    cmd.terminate()
    assert_equal(cmd.join(), [0])


# Calling terminate on an already joined command is acceptable ...
@with_temp_folder
def test_atomiccmd__terminate_after_join(temp_folder):
    cmd = AtomicCmd("true")
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])
    cmd.terminate()
    assert_equal(cmd.join(), [0])


# Signals are translated into strings
@with_temp_folder
def test_atomiccmd__terminate_sigterm(temp_folder):
    cmd = AtomicCmd(("sleep", "10"))
    cmd.run(temp_folder)
    cmd.terminate()
    assert_equal(cmd.join(), ["SIGTERM"])


@with_temp_folder
def test_atomiccmd__terminate_sigkill(temp_folder):
    cmd = AtomicCmd(("sleep", "10"))
    cmd.run(temp_folder)
    cmd._proc.kill()
    assert_equal(cmd.join(), ["SIGKILL"])


###############################################################################
###############################################################################
# commit

def _setup_for_commit(temp_folder, create_cmd=True):
    destination = os.path.join(temp_folder, "out")
    temp_folder = os.path.join(temp_folder, "tmp")
    os.makedirs(destination)
    os.makedirs(temp_folder)

    if not create_cmd:
        return destination, temp_folder

    cmd = AtomicCmd(("touch", "%(OUT_FOO)s"),
                    OUT_FOO=os.path.join(destination, "1234"))
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])

    return destination, temp_folder, cmd


@with_temp_folder
def test_atomiccmd__commit_simple(temp_folder):
    destination, temp_folder, cmd = _setup_for_commit(temp_folder)
    cmd.commit(temp_folder)
    assert not os.path.exists(os.path.join(temp_folder, "1234"))
    assert os.path.exists(os.path.join(destination, "1234"))


@with_temp_folder
def test_atomiccmd__commit_temp_out(temp_folder):
    dest, temp = _setup_for_commit(temp_folder, create_cmd=False)
    cmd = AtomicCmd(("echo", "foo"),
                    OUT_STDOUT=os.path.join(dest, "foo.txt"),
                    TEMP_OUT_FOO="bar.txt")
    cmd.run(temp)
    assert_equal(cmd.join(), [0])
    set_file_contents(os.path.join(temp, "bar.txt"), "1 2 3")
    cmd.commit(temp)
    assert_equal(os.listdir(temp), [])
    assert_equal(os.listdir(dest), ["foo.txt"])


@with_temp_folder
def test_atomiccmd__commit_temp_only(temp_folder):
    cmd = AtomicCmd(("echo", "foo"),
                    TEMP_OUT_STDOUT="bar.txt")
    cmd.run(temp_folder)
    assert_equal(cmd.join(), [0])
    assert os.path.exists(os.path.join(temp_folder, "bar.txt"))
    cmd.commit(temp_folder)
    assert_equal(os.listdir(temp_folder), [])


def test_atomiccmd__commit_before_run():
    cmd = AtomicCmd("true")
    assert_raises(CmdError, cmd.commit, "/tmp")


@with_temp_folder
def test_atomiccmd__commit_while_running(temp_folder):
    cmd = AtomicCmd(("sleep", "10"))
    cmd.run(temp_folder)
    assert_raises(CmdError, cmd.commit, temp_folder)
    cmd.terminate()
    cmd.join()


@with_temp_folder
def test_atomiccmd__commit_before_join(temp_folder):
    cmd = AtomicCmd(("sleep", "0.1"))
    cmd.run(temp_folder)
    while cmd._proc.poll() is None:
        pass
    assert_raises(CmdError, cmd.commit, temp_folder)
    cmd.join()


# The temp path might differ, as long as the actual path is the same
@with_temp_folder
def test_atomiccmd__commit_temp_folder(temp_folder):
    destination, temp_folder, cmd = _setup_for_commit(temp_folder)
    cmd.commit(os.path.realpath(temp_folder))
    assert not os.path.exists(os.path.join(temp_folder, "1234"))
    assert os.path.exists(os.path.join(destination, "1234"))


@with_temp_folder
def test_atomiccmd__commit_wrong_temp_folder(temp_folder):
    destination, temp_folder, cmd = _setup_for_commit(temp_folder)
    assert_raises(CmdError, cmd.commit, destination)


@with_temp_folder
def test_atomiccmd__commit_missing_files(temp_folder):
    destination, temp_folder = _setup_for_commit(temp_folder, False)
    cmd = AtomicCmd(("touch", "%(OUT_FOO)s"),
                    OUT_FOO=os.path.join(destination, "1234"),
                    OUT_BAR=os.path.join(destination, "4567"))
    cmd.run(temp_folder)
    cmd.join()
    before = set(os.listdir(temp_folder))
    assert_raises(CmdError, cmd.commit, temp_folder)
    assert_equal(before, set(os.listdir(temp_folder)))


@with_temp_folder
def test_atomiccmd__commit_failure_cleanup(temp_folder):
    counter = []
    move_file = fileutils.move_file

    def _monkey_move_file(source, destination):
        if counter:
            raise OSError("ARRRGHHH!")
        counter.append(destination)

        return move_file(source, destination)

    destination, temp_folder = _setup_for_commit(temp_folder, False)
    command = AtomicCmd(("touch", "%(OUT_FILE_1)s", "%(OUT_FILE_2)s",
                         "%(OUT_FILE_3)s"),
                        OUT_FILE_1=os.path.join(destination, "file_1"),
                        OUT_FILE_2=os.path.join(destination, "file_2"),
                        OUT_FILE_3=os.path.join(destination, "file_3"))

    try:
        fileutils.move_file = _monkey_move_file
        command.run(temp_folder)
        assert_equal(command.join(), [0])
        assert_raises(OSError, command.commit, temp_folder)

        assert_equal(tuple(os.listdir(destination)), ())
    finally:
        fileutils.move_file = move_file


@with_temp_folder
def test_atomiccmd__commit_with_pipes(temp_folder):
    destination, temp_folder = _setup_for_commit(temp_folder, False)
    command_1 = AtomicCmd(("echo", "Hello, World!"),
                          OUT_STDOUT=AtomicCmd.PIPE)
    command_2 = AtomicCmd(("gzip",),
                          IN_STDIN=command_1,
                          OUT_STDOUT=os.path.join(destination, "foo.gz"))

    command_1.run(temp_folder)
    command_2.run(temp_folder)

    assert_equal(command_1.join(), [0])
    assert_equal(command_2.join(), [0])

    command_1.commit(temp_folder)
    command_2.commit(temp_folder)

    assert_equal(set(os.listdir(destination)), set(("foo.gz",)))
    assert_equal(set(os.listdir(temp_folder)), set())


###############################################################################
###############################################################################
# __str__
# Additional tests in atomicpp_test.py

def test_atomiccmd__str__():
    cmd = AtomicCmd(("echo", "test"))
    assert_equal(paleomix.atomiccmd.pprint.pformat(cmd), str(cmd))


###############################################################################
###############################################################################
# Cleanup

# Test that the internal list of processes is kept clean of old objects
def test_atomiccmd__cleanup_proc():
    @with_temp_folder
    def _do_test_atomiccmd__cleanup_proc(temp_folder, func):
        assert_equal(paleomix.atomiccmd.command._PROCS, set())
        cmd = AtomicCmd("ls")
        cmd.run(temp_folder)
        ref = iter(paleomix.atomiccmd.command._PROCS).next()
        assert ref
        assert_equal(ref(), cmd._proc)

        assert_equal(cmd.join(), [0])
        cmd = func(cmd, temp_folder)

        assert ref not in paleomix.atomiccmd.command._PROCS

    def _do_commit(cmd, temp_folder):
        # Trigger freeing of proc
        cmd.commit(temp_folder)
        return cmd

    # The proc object should be released when commit is called
    yield _do_test_atomiccmd__cleanup_proc, _do_commit
    # The proc object should be released when the cmd object is released
    yield _do_test_atomiccmd__cleanup_proc, lambda _cmd, _temp_folder: None


def test_atomiccmd__cleanup_sigterm():
    def _do_test_atomiccmd__cleanup_sigterm(kill_at):
        sigs_sent, exit_called = {}, []

        def _wrap_killpg(pid, sig):
            assert pid not in sigs_sent
            do_kill = len(sigs_sent) == kill_at
            sigs_sent[pid] = (sig, do_kill)

            # Simulate already terminated processes; cleanup should continue
            if do_kill:
                raise OSError("KABOOM!")

        def _wrap_exit(returncode):
            exit_called.append(returncode)

        _procs = [flexmock(pid=7913),
                  # I've got the same combination on my luggage!
                  flexmock(pid=12345)]

        assert not paleomix.atomiccmd.command._PROCS
        with Monkeypatch("paleomix.atomiccmd.command._PROCS", _procs):
            assert_equal(len(paleomix.atomiccmd.command._PROCS), 2)
            with Monkeypatch("os.killpg", _wrap_killpg):
                with Monkeypatch("sys.exit", _wrap_exit):
                    paleomix.atomiccmd.command._cleanup_children(
                        signal.SIGTERM, None)

        assert_equal(exit_called, [-signal.SIGTERM])
        assert_equal(sigs_sent, {7913: (signal.SIGTERM, kill_at == 0),
                                 12345: (signal.SIGTERM, kill_at == 1)})

    yield _do_test_atomiccmd__cleanup_sigterm, 0
    yield _do_test_atomiccmd__cleanup_sigterm, 1
    yield _do_test_atomiccmd__cleanup_sigterm, 2


# Ensure that the cleanup function handles weakrefs that have been freed
def test_atomiccmd__cleanup_sigterm__dead_weakrefs():
    exit_called = []
    procs_wrapper = [weakref.ref(Monkeypatch("sys.exit", None))]

    assert_equal(procs_wrapper[0](), None)

    def _wrap_killpg(_pid, _sig):
        assert False  # pragma: no coverage

    def _wrap_exit(returncode):
        exit_called.append(returncode)

    with Monkeypatch("paleomix.atomiccmd.command._PROCS", procs_wrapper):
        with Monkeypatch("os.killpg", _wrap_killpg):
            with Monkeypatch("sys.exit", _wrap_exit):
                paleomix.atomiccmd.command._cleanup_children(
                    signal.SIGTERM, None)
    assert_equal(exit_called, [-signal.SIGTERM])



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Wrapper around "samtools mpileup | bcftools view", to improve performance
when genotyping sparse regions (e.g. sets of exons / RNAs / or similar), and
allow transparent multi-threading. Alternatively, only "samtools mpileup" may
be called, in order to generate pileups for a set of regions.

There are 3 main motivations for this wrapper:
    1. Current versions of SAMTools read the full contents of the input BAM,
       when a BED file of regions is specified, even if these regions cover
       only a fraction of sites. This can be somewhat mitigated by ALSO
       specifying a region using -r, which fetches just that region, but this
       does not scale well for thousands of individual regions.
    2. It provides transparent parallelization, allowing any set of bed regions
       to be split and processed in parallel.
"""
import os
import sys
import shutil
import signal
import argparse
import traceback
import multiprocessing

import pysam

from paleomix.common.bedtools import \
    read_bed_file, \
    sort_bed_by_bamfile

import paleomix.tools.factory as factory
import paleomix.common.procs as processes


class BatchError(RuntimeError):
    pass


# Size of smallest block in (linear) BAM index (= 2 << 14)
_BAM_BLOCK_SIZE = 16384


###############################################################################
###############################################################################
# CLI functions

def build_call(call, args, positional, new_args):
    call = list(call)
    args = dict(args)

    for new_arg in new_args:
        key, value = new_arg, None
        if "=" in new_arg:
            key, value = new_arg.split("=", 1)
        args[key] = value

    for (key, value) in sorted(args.iteritems()):
        call.append(key)
        if value is not None:
            call.append(value)

    call.extend(positional)

    return call


###############################################################################
###############################################################################
# BAM filtering mode

def filter_bam(bamfile, bedfile):
    with pysam.Samfile(bamfile) as bam_handle_in:
        regions = collect_regions(bedfile, bam_handle_in)
        regions.reverse()

        with pysam.Samfile("-", "wbu",
                           template=bam_handle_in) as bam_handle_out:
            while regions:
                region_aend = 0
                contig, start, end = regions[-1]
                for record in bam_handle_in.fetch(contig, start):
                    current_aend = record.aend
                    region_aend = max(region_aend, current_aend)
                    if record.pos > end:
                        last_contig, _, _ = regions.pop()
                        if not regions:
                            break

                        contig, start, end = regions[-1]
                        if (region_aend + _BAM_BLOCK_SIZE < start) \
                                or (contig != last_contig):
                            break

                    if current_aend >= start:
                        bam_handle_out.write(record)
                else:  # Reached the end of this contig
                    while regions and (regions[-1][0] == contig):
                        regions.pop()

    return 0


###############################################################################
###############################################################################
# Common functions

def cleanup_batch(setup):
    sys.stderr.write("Cleaning up batch ...\n")
    for handle in setup["handles"].itervalues():
        handle.close()

    for proc in setup["procs"].itervalues():
        if proc.poll() is None:
            proc.terminate()
            proc.wait()

    for filename in setup["temp_files"].itervalues():
        sys.stderr.write("Removing temporary file %r\n" % (filename,))
        os.remove(filename)


def write_bed_file(prefix, regions):
    fpath = prefix + ".bed"
    with open(fpath, "w") as bed_handle:
        for (contig, start, end) in regions:
            bed_handle.write("%s\t%i\t%i\n" % (contig, start, end))
        bed_handle.flush()
    return fpath


def setup_basic_batch(args, regions, prefix, func, first_batch=True):
    setup = {"files": {},
             "temp_files": {},
             "procs": {},
             "handles": {}}

    try:
        setup["files"]["bed"] = write_bed_file(prefix, regions)
        setup["temp_files"]["bed"] = setup["files"]["bed"]

        filter_builder = factory.new("genotype")
        filter_builder.set_option("--filter-only")
        filter_builder.set_option("--bedfile", setup["files"]["bed"])
        filter_builder.add_option(args.bamfile)
        filter_builder.add_option(args.destination)

        setup["procs"]["filter"] \
            = processes.open_proc(filter_builder.call,
                                  stdout=processes.PIPE)

        call_stdout = func(setup)
        if not first_batch:
            setup["procs"]["grep"] = processes.open_proc(('grep', '-v', '^#'),
                                                         stdin=call_stdout,
                                                         stdout=processes.PIPE)
            call_stdout = setup["procs"]["grep"].stdout

        setup["handles"]["outfile"] = open(prefix, "w")
        zip_proc = processes.open_proc(["bgzip"],
                                       stdin=call_stdout,
                                       stdout=setup["handles"]["outfile"])

        setup["procs"]["gzip"] = zip_proc

        return setup
    except:
        sys.stderr.write(traceback.format_exc() + "\n")
        cleanup_batch(setup)
        raise


###############################################################################
###############################################################################
# Pileup batch generation

def setup_mpileup_batch(args, regions, prefix, first_batch=True):
    def _create_mpileup_proc(setup):
        mpileup_args = {"-l": setup["files"]["bed"]}
        call = build_call(call=("samtools", "mpileup"),
                          args=mpileup_args,
                          new_args=args.mpileup_argument,
                          positional=("-",))

        sys.stderr.write("Running 'samtools mpileup': %s\n" % (" ".join(call)))
        procs = setup["procs"]
        procs["mpileup"] \
            = processes.open_proc(call,
                                  stdin=procs["filter"].stdout,
                                  stdout=processes.PIPE)

        return procs["mpileup"].stdout

    return setup_basic_batch(args, regions, prefix, _create_mpileup_proc,
                             first_batch=first_batch)


###############################################################################
###############################################################################
# Genotyping batch generation

def setup_genotyping_batch(args, regions, prefix, first_batch=True):
    def _create_genotyping_proc(setup):
        mpileup_args = {"-u": None,
                        "-l": setup["files"]["bed"]}
        mpileup_call = build_call(call=("samtools", "mpileup"),
                                  args=mpileup_args,
                                  new_args=args.mpileup_argument,
                                  positional=("-",))

        sys.stderr.write("Running 'samtools mpileup': %s\n"
                         % (" ".join(mpileup_call)))

        procs = setup["procs"]
        procs["mpileup"] \
            = processes.open_proc(mpileup_call,
                                  stdin=procs["filter"].stdout,
                                  stdout=processes.PIPE)

        bcftools_call = build_call(call=("bcftools", "view"),
                                   args={},
                                   new_args=args.bcftools_argument,
                                   positional=("-",))

        sys.stderr.write("Running 'bcftools call': %s\n"
                         % (" ".join(bcftools_call)))

        procs["bcftools"] \
            = processes.open_proc(bcftools_call,
                                  stdin=procs["mpileup"].stdout,
                                  stdout=processes.PIPE)

        return procs["bcftools"].stdout

    return setup_basic_batch(args, regions, prefix, _create_genotyping_proc,
                             first_batch=first_batch)


###############################################################################
###############################################################################

def setup_batch(args, regions, filename, first_batch):
    """Setup a batch; either a full genotyping, or just a pileup depending on
    'args.pileup_only'; the results are written to 'filename'.
    """
    if args.pileup_only:
        return setup_mpileup_batch(args, regions, filename, first_batch)
    return setup_genotyping_batch(args, regions, filename, first_batch)


def run_batch((args, regions, filename, first_batch)):
    setup = setup_batch(args, regions, filename, first_batch)
    try:
        if any(processes.join_procs(setup["procs"].values())):
            return None

        return filename
    except:
        # Re-wrap exception with full-traceback; otherwise this information
        # is lost when the exception is retrieved in the main process.
        raise BatchError(traceback.format_exc())
    finally:
        cleanup_batch(setup)


###############################################################################
###############################################################################

def init_worker_thread():
    """Init function for subprocesses created by multiprocessing.Pool: Ensures
    that KeyboardInterrupts only occur in the main process, allowing us to do
    proper cleanup.
    """
    signal.signal(signal.SIGINT, signal.SIG_IGN)


###############################################################################
###############################################################################

def merge_bed_regions(regions):
    """Takes a sequence of bed regions [(contig, start, end), ...], which is
    assumed to be sorted by contig and coordiates, and returns a list in which
    overlapping records are merged into one larger region.
    """
    merged = []
    last_contig = last_start = last_end = None
    for record in regions:
        if (record.contig != last_contig) or (record.start > last_end):
            if last_contig is not None:
                merged.append((last_contig, last_start, last_end))
            last_contig = record.contig
            last_start = record.start
            last_end = record.end
        else:
            last_start = min(last_start or 0, record.start)
            last_end = max(last_end, record.end)

    if last_contig is not None:
        merged.append((last_contig, last_start, last_end))

    return merged


def create_batches(args, regions):
    """Yields a sequence of batches that may be passed to the 'run_batch'
    function; each batch consists of the 'args' object, a set of BED regions,
    and a destination filename. The set of BED regions is derived by splitting
    the total set of regions into args.nbatches portions.
    """
    tmpl = "{0}.batch_%03i".format(args.destination)

    def _get_batch_fname(count):
        """Returns a filename for batch number 'count'."""
        if count:
            return tmpl % (count,)
        return args.destination

    total_size = sum(end - start for (_, start, end) in regions)
    batch_size = total_size // args.nbatches + 5

    batch_count = 0
    current_batch = []
    current_total = 0
    for (contig, start, end) in regions:
        while (end - start) + current_total > batch_size:
            new_end = start + batch_size - current_total
            current_batch.append((contig, start, new_end))
            start = new_end
            yield args, current_batch, _get_batch_fname(batch_count), \
                not batch_count
            current_batch = []
            current_total = 0
            batch_count += 1
        current_batch.append((contig, start, end))
        current_total += end - start

    if current_batch:
        yield args, current_batch, _get_batch_fname(batch_count), \
            not batch_count


def merge_batch_results(filenames_iter):
    """Takes a multiprocessing.imap iterator yielding filenames of completed
    batches (gzipped vcf or mpileup files), and writes these into the
    file-handle out.
    """
    while True:
        try:
            # A timeout allows iteruption by the user, which is not the
            # case otherwise. The value is arbitrary.
            target_filename = filenames_iter.next(60)
            # None signals error in subprocess; see 'run_batch'
            if target_filename is None:
                return False

            sys.stderr.write("Merging into file: %r\n" % (target_filename,))
            break
        except multiprocessing.TimeoutError:
            pass
        except StopIteration:
            return

    with open(target_filename, "r+") as target_handle:
        while True:
            try:
                filename = filenames_iter.next(60)
                sys.stderr.write("   - Processing batch: %r" % (filename,))

                # BGZip is terminated by 28b empty block (cf. ref)
                # While the standard implies that these should be ignored
                # if not actually at the end of the file, the tabix tool
                # stops processing at the first such block it encounters
                target_handle.seek(-28, 2)
                with open(filename) as input_handle:
                    shutil.copyfileobj(input_handle, target_handle)
                os.remove(filename)
            except multiprocessing.TimeoutError:
                pass
            except StopIteration:
                break

    return True


def collect_regions(bedfile, bam_input_handle):
    """Returns the regions to be genotyped / pileup'd, as a list of bed-regions
    in the form (contig, start, end), where start is zero-based, and end is
    open based.
    """
    if bedfile is not None:
        regions = list(read_bed_file(bedfile))
        sort_bed_by_bamfile(bam_input_handle, regions)
        regions = merge_bed_regions(regions)
    else:
        regions = []
        for (name, length) in zip(bam_input_handle.references,
                                  bam_input_handle.lengths):
            regions.append((name, 0, length))
    return regions


def process_batches(args, batches):
    """Runs a set of batches, and merges the resulting output files if more
    than one batch is included.
    """
    nbatches = min(args.nbatches, len(batches))
    pool = multiprocessing.Pool(nbatches, init_worker_thread)

    try:
        batches = pool.imap(run_batch, batches, 1)
        if not merge_batch_results(batches):
            pool.terminate()
            pool.join()
            return 1

        pool.close()
        pool.join()
        return 0
    except:
        pool.terminate()
        pool.join()
        raise


def create_empty_bgz(destination):
    """Writes an empty BGZip file to the given destination; this file contains
    a single empty BGZip block (28b).
    """
    with open(destination, "w") as output:
        # Empty BGZip block
        output.write("\x1f\x8b\x08\x04\x00\x00\x00\x00\x00\xff\x06\x00\x42")
        output.write("\x43\x02\x00\x1b\x00\x03\x00\x00\x00\x00\x00\x00\x00")
        output.write("\x00\x00")


def parse_args(argv):
    prog = "paleomix genotype"
    usage = "%s [options] sorted.bam out.vcf.bgz" % (prog,)

    parser = argparse.ArgumentParser(prog=prog, usage=usage)
    parser.add_argument("bamfile", metavar='INPUT',
                        help="Sorted and indexed BAM file.")
    parser.add_argument("destination", metavar='OUTPUT',
                        help="BGZip compressed VCF or pileup. Also used as "
                             "prefix for temporary files.")
    parser.add_argument('--bedfile', default=None, metavar="BED",
                        help="Optional bedfile, specifying regions to pileup "
                             "or genotype [Default: %(default)s].")
    parser.add_argument('--mpileup-argument', default=[], action="append",
                        help="Pass argument to 'samtools mpileup'; must be "
                             "used as follows: --mpileup-argument=-argument "
                             "for arguments without values, and "
                             "--mpileup-argument=-argument=value for "
                             "arguments with values.")
    parser.add_argument('--bcftools-argument', default=[], action="append",
                        help="Pass argument to 'bcftools view'; see the "
                             "--mpileup-argument command description.")
    parser.add_argument('--pileup-only', default=False, action="store_true",
                        help="Only run 'samtools mpileup', generating a text "
                             "pileup instead of a VCF file [Default: off].")
    parser.add_argument('--nbatches', metavar="N", default=1, type=int,
                        help="Split the BED into N number of batches, which "
                             "are run in parallel [Default: %(default)s].")
    parser.add_argument('--overwrite', default=False, action="store_true",
                        help="Overwrite output if it already exists "
                             "[Default: no].")

    # When set, the --bedfile argument is read and used to filter the BAM
    # specified for the 'bamfile' parameter; all other parameters are ignored.
    parser.add_argument('--filter-only', default=False, action="store_true",
                        help=argparse.SUPPRESS)

    return parser.parse_args(argv)


def main(argv):
    args = parse_args(argv)
    if args.filter_only:
        if not args.bedfile:
            sys.stderr.write("--filter-only requires --bedfile; terminating\n")
            return 1

        return filter_bam(args.bamfile, args.bedfile)

    if os.path.exists(args.destination) and not args.overwrite:
        sys.stderr.write("Output already exists; use --overwrite to allow "
                         "overwriting of this file.\n")
        return 1

    with pysam.Samfile(args.bamfile) as bam_input_handle:
        regions = collect_regions(args.bedfile, bam_input_handle)
    batches = list(create_batches(args, regions))
    if not batches:
        create_empty_bgz(args.destination)
        return 0

    try:
        return process_batches(args, batches)
    except BatchError, error:
        sys.stderr.write("ERROR while processing BAM:\n")
        sys.stderr.write("    %s\n"
                         % ("\n    ".join(str(error).split("\n"),)))
        return 1

    return 0



#!/usr/bin/python
#
# Copyright (c) 2016 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import logging
import os
import shutil
import tarfile
import time

import paleomix
import paleomix.yaml
import paleomix.logger

import paleomix.common.fileutils as fileutils

from paleomix.common.console import \
    print_err, \
    print_info, \
    print_warn

from paleomix.common.formats.fasta import FASTA

from paleomix.pipeline import \
    Pypeline

from paleomix.nodes.samtools import \
    BAMIndexNode

from paleomix.nodes.raxml import \
    RAxMLRapidBSNode

import paleomix.tools.bam_pipeline.mkfile as bam_mkfile

import paleomix.tools.zonkey.config as zonkey_config
import paleomix.tools.zonkey.parts.mitochondria as mitochondria
import paleomix.tools.zonkey.parts.nuclear as nuclear
import paleomix.tools.zonkey.parts.report as report
import paleomix.tools.zonkey.parts.summary as summary
import paleomix.tools.zonkey.parts.common as common_nodes


def run_pipeline(config, nodes, msg):
    pipeline = Pypeline(config)
    pipeline.add_nodes(nodes)

    logfile_template = time.strftime("zonkey_pipeline.%Y%m%d_%H%M%S_%%02i.log")
    paleomix.logger.initialize(config, logfile_template)

    logger = logging.getLogger(__name__)
    logger.info(msg)

    if config.list_executables:
        pipeline.print_required_executables()
        return True
    elif config.list_output_files:
        pipeline.print_output_files()
        return True
    elif config.list_input_files:
        pipeline.print_input_files()
        return True
    elif config.dot_file:
        logger.info("Writing dependency graph to %r ...", config.dot_file)
        return pipeline.to_dot(config.dot_file)

    return pipeline.run(max_threads=config.max_threads,
                        progress_ui=config.progress_ui,
                        dry_run=config.dry_run)


def build_plink_nodes(config, data, root, bamfile, dependencies=()):
    plink = {"root": os.path.join(root, 'results', 'plink')}

    ped_node = nuclear.BuildTPEDFilesNode(output_root=plink["root"],
                                          table=config.tablefile,
                                          downsample=config.downsample_to,
                                          bamfile=bamfile,
                                          dependencies=dependencies)

    for postfix in ('incl_ts', 'excl_ts'):
        parameters = {
            "output_prefix": os.path.join(plink["root"], postfix),
            "tfam": os.path.join(plink["root"], "common.tfam"),
            "tped": os.path.join(plink["root"], postfix + ".tped"),
            "plink_parameters": config.database.settings["Plink"],
            "dependencies": (ped_node,),
        }

        if config.indep:
            parameters["indep_filter"] = config.indep
            parameters["indep_parameters"] = config.indep_params

            bed_node = nuclear.BuildFilteredBEDFilesNode(**parameters)
        else:
            bed_node = nuclear.BuildBEDFilesNode(**parameters)

        plink[postfix] = bed_node

    return plink


def build_admixture_nodes(config, data, root, plink):
    nodes = []
    for postfix in ('incl_ts', 'excl_ts'):
        bed_node = plink[postfix]

        admix_root = os.path.join(root, "results", "admixture")
        report_root = os.path.join(root, "figures", "admixture")
        for k_groups in (2, 3):
            replicates = []

            input_file = os.path.join(plink["root"], postfix + ".bed")
            for replicate in xrange(config.admixture_replicates):
                output_root = os.path.join(admix_root, "%02i" % (replicate,))

                node = nuclear.AdmixtureNode(input_file=input_file,
                                             output_root=output_root,
                                             k_groups=k_groups,
                                             samples=data.samples,
                                             dependencies=(bed_node,))

                replicates.append(node)

            node = nuclear.SelectBestAdmixtureNode(replicates=replicates,
                                                   output_root=admix_root)

            if config.admixture_only:
                nodes.append(node)
            else:
                samples = os.path.join(root, "figures", "samples.txt")
                plot = nuclear.AdmixturePlotNode(input_file=os.path.join(admix_root, "%s.%i.Q" % (postfix, k_groups)),
                                                 output_prefix=os.path.join(report_root, "%s_k%i" % (postfix, k_groups)),
                                                 samples=samples,
                                                 order=data.sample_order,
                                                 dependencies=node)

                nodes.append(plot)

    return nodes


def build_treemix_nodes(config, data, root, plink):
    tmix_root = os.path.join(root, 'results', 'treemix')

    nodes = []
    for postfix in ('incl_ts', 'excl_ts'):
        plink_prefix = os.path.join(plink["root"], postfix)
        plink_nodes = plink[postfix]

        freq_node = nuclear.BuildFreqFilesNode(output_prefix=plink_prefix,
                                               input_prefix=os.path.join(plink["root"], postfix),
                                               tfam=os.path.join(plink["root"], "common.tfam"),
                                               parameters=config.database.settings["Plink"],
                                               dependencies=plink_nodes)

        tmix_prefix = os.path.join(tmix_root, postfix)
        tmix_file_node = nuclear.FreqToTreemixNode(input_file=plink_prefix + ".frq.strat.gz",
                                                   output_file=tmix_prefix + ".gz",
                                                   dependencies=(freq_node,))

        k_snps = config.treemix_k
        if not k_snps:
            k_snps = ('n_sites_%s' % (postfix,),
                      os.path.join(plink["root"], "common.summary"))

        for n_migrations in (0, 1):
            n_prefix = "%s.%i" % (tmix_prefix, n_migrations)

            tmix_node = nuclear.TreemixNode(data=data,
                                            input_file=tmix_prefix + ".gz",
                                            output_prefix=n_prefix,
                                            m=n_migrations,
                                            k=k_snps,
                                            outgroup=config.treemix_outgroup,
                                            dependencies=(tmix_file_node,))

            samples = os.path.join(root, "figures", "samples.txt")
            output_prefix = os.path.join(root, "figures", "treemix", "%s_%i" % (postfix, n_migrations))
            plot_node = nuclear.PlotTreemixNode(samples=samples,
                                                prefix=n_prefix,
                                                output_prefix=output_prefix,
                                                dependencies=(tmix_node,))

            nodes.append(plot_node)

    return nodes


def build_pca_nodes(config, data, root, plink):
    pca_root = os.path.join(root, 'results', 'pca')

    nodes = []
    for postfix in ('incl_ts', 'excl_ts'):
        plink_prefix = os.path.join(plink["root"], postfix)
        plink_nodes = plink[postfix]

        pca_prefix = os.path.join(pca_root, postfix)
        pca_node = nuclear.SmartPCANode(input_prefix=plink_prefix,
                                        output_prefix=pca_prefix,
                                        nchroms=data.settings["NChroms"],
                                        dependencies=plink_nodes)

        samples = os.path.join(root, "figures", "samples.txt")
        pca_plots = os.path.join(root, "figures", "pca", postfix)
        pca_plot_node = nuclear.PlotPCANode(samples=samples,
                                            prefix=pca_prefix,
                                            output_prefix=pca_plots,
                                            dependencies=pca_node)

        nodes.append(pca_plot_node)

    return nodes


def build_coverage_nodes(data, root, nuc_bam, dependencies=()):
    output_prefix = os.path.join(root, 'figures', 'coverage', 'coverage')

    return nuclear.PlotCoverageNode(contigs=data.contigs,
                                    input_file=nuc_bam,
                                    output_prefix=output_prefix,
                                    dependencies=dependencies),


def build_mito_nodes(config, root, bamfile, dependencies=()):
    if config.database.mitochondria is None:
        print_warn("WARNING: Zonkey database %r does not contain "
                   "mitochondrial  sequences; cannot analyze MT BAM %r!\n"
                   % (config.tablefile, bamfile))
        return ()

    samples = os.path.join(root, "figures", "samples.txt")

    mt_prefix = os.path.join(root, "results", "mitochondria", "sequences")
    alignment = mitochondria.MitoConsensusNode(database=config.tablefile,
                                               bamfile=bamfile,
                                               output_prefix=mt_prefix,
                                               dependencies=dependencies)

    raxml_template = os.path.join(root, "results", "mitochondria", "raxml_%s")
    phylo = RAxMLRapidBSNode.customize(input_alignment=mt_prefix + ".phy",
                                       output_template=raxml_template,
                                       dependencies=(alignment,))

    phylo.command.set_option("-N", 100)
    phylo.command.set_option("-m", "GTRGAMMA")
    phylo = phylo.build_node()

    output_prefix = os.path.join(root, "figures", "mitochondria", "mito_phylo")
    trees = mitochondria.DrawPhylogenyNode(samples=samples,
                                           treefile=raxml_template % ("bestTree",),
                                           bootstraps=raxml_template % ("bootstrap",),
                                           output_prefix=output_prefix,
                                           dependencies=(phylo,))

    return (trees,)


def build_pipeline(config, root, nuc_bam, mito_bam, cache):
    nodes = []
    sample_tbl = os.path.join(root, "figures", "samples.txt")
    samples = common_nodes.WriteSampleList(config=config,
                                           output_file=sample_tbl)

    if nuc_bam is not None:
        # When not sampling, BuildTPED relies on indexed access to ease
        # processing of one chromosome at a time. The index is further required
        # for idxstats used by the PlotCoverageNode.
        index = cache.get(nuc_bam)
        if index is None:
            index = cache[nuc_bam] = BAMIndexNode(infile=nuc_bam)

        plink = build_plink_nodes(config, config.database, root, nuc_bam,
                                  dependencies=(samples, index))

        nodes.extend(build_admixture_nodes(config, config.database, root,
                                           plink))

        if not config.admixture_only:
            nodes.extend(build_coverage_nodes(config.database,
                                              root, nuc_bam, (index,)))
            nodes.extend(build_pca_nodes(config, config.database,
                                         root, plink))
            nodes.extend(build_treemix_nodes(config, config.database,
                                             root, plink))

    if mito_bam is not None and not config.admixture_only:
        index = cache.get(mito_bam)
        if index is None:
            index = cache[mito_bam] = BAMIndexNode(infile=mito_bam)

        nodes.extend(build_mito_nodes(config, root, mito_bam,
                                      dependencies=(samples, index)))

    if not config.admixture_only:
        nodes.append(report.ReportNode(config, root, nuc_bam, mito_bam,
                                       dependencies=nodes))

    return nodes


def run_admix_pipeline(config):
    print_info("\nBuilding %i Zonkey pipeline(s):" % (len(config.samples),))
    config.temp_root = os.path.join(config.destination, "temp")
    if not config.dry_run:
        fileutils.make_dirs(config.temp_root)

    cache = {}
    nodes = []
    items = config.samples.iteritems()
    for idx, (name, sample) in enumerate(sorted(items), start=1):
        root = sample["Root"]
        nuc_bam = sample["Files"].get("Nuc")
        mito_bam = sample["Files"].get("Mito")

        genomes = []
        if mito_bam:
            genomes.append("MT")
        if nuc_bam:
            genomes.append("Nuclear")

        print_info("  %i. %s: %s DNA" % (idx, name, ' and '.join(genomes)))

        nodes.extend(build_pipeline(config, root, nuc_bam, mito_bam, cache))

    if config.multisample and not config.admixture_only:
        nodes = [summary.SummaryNode(config, nodes)]

    if not run_pipeline(config, nodes, "\nRunning Zonkey:"):
        return 1


def setup_mito_mapping(config):
    genomes_root = os.path.join(config.destination, "genomes")
    if not os.path.exists(genomes_root):
        fileutils.make_dirs(genomes_root)

    mkfile_fpath = os.path.join(config.destination, "makefile.yaml")

    filenames = [mkfile_fpath]
    for name, record in sorted(config.database.mitochondria.iteritems()):
        filenames.append(os.path.join(genomes_root, "%s.fasta"
                                      % (record.name,)))

    existing_filenames = [filename for filename in filenames
                          if os.path.exists(filename)]

    # A bit strict, but avoid accidential overwrites
    if existing_filenames:
        print_err("ERROR: Output file(s) already exists, "
                  "cannot proceed:\n    %s"
                  % ("\n    ".join(map(repr, existing_filenames),)))

        return 1

    with open(mkfile_fpath, "w") as mkfile:
        mkfile.write(bam_mkfile.build_makefile(add_prefix_tmpl=False,
                                               add_sample_tmpl=False))

        mkfile.write("\n\nPrefixes:\n")

        for name, record in sorted(config.database.mitochondria.iteritems()):
            meta = (record.meta or "").upper()
            if "EXCLUDE" in meta:
                continue

            mkfile.write("  %s:\n" % (record.name,))
            mkfile.write("    Path: genomes/%s.fasta\n" % (record.name,))

            info = config.database.samples.get(record.name)
            if info is not None:
                mkfile.write("    # Group: %s\n"
                             % (info.get('Group(3)', 'NA'),))
                mkfile.write("    # Species: %s\n"
                             % (info.get('Species', 'NA'),))
                mkfile.write("    # Sex: %s\n"
                             % (info.get('Sex', 'NA'),))
                mkfile.write("    # Publication: %s\n"
                             % (info.get('Publication', 'NA'),))
                mkfile.write("    # Sample ID: %s\n"
                             % (info.get('SampleID', 'NA'),))

            mkfile.write('\n')

            fasta_fpath = os.path.join(genomes_root,
                                       "%s.fasta" % (record.name,))

            with open(fasta_fpath, "w") as fasta_handle:
                record = FASTA(
                    name=record.name,
                    meta=None,
                    sequence=record.sequence.replace('-', ''))

                fasta_handle.write(str(record))
                fasta_handle.write("\n")

        mkfile.write("\n")

    return 0


def setup_example(config):
    root = os.path.join(config.destination, 'zonkey_pipeline')

    with tarfile.TarFile(config.tablefile) as tar_handle:
        example_files = []
        existing_files = []
        for member in tar_handle.getmembers():
            if os.path.dirname(member.name) == 'examples' and member.isfile():
                example_files.append(member)

                destination = fileutils.reroot_path(root, member.name)
                if os.path.exists(destination):
                    existing_files.append(destination)

        if existing_files:
            print_err("Output files already exist at destination:\n    - %s"
                      % ("\n    - ".join(map(repr, existing_files))))
            return 1
        elif not example_files:
            print_err("Sample database %r does not contain example data; "
                      "cannot proceed." % (config.tablefile,))
            return 1

        if not os.path.exists(root):
            fileutils.make_dirs(root)

        for member in example_files:
            destination = fileutils.reroot_path(root, member.name)
            src_handle = tar_handle.extractfile(member)
            with open(destination, 'w') as out_handle:
                shutil.copyfileobj(src_handle, out_handle)

    print_info("Sucessfully saved example data in %r" % (root,))

    return 0


def main(argv):
    try:
        config = zonkey_config.parse_config(argv)
        if config is None:
            return 1
    except zonkey_config.ConfigError, error:
        print_err(error)
        return 1

    if config.command == "run":
        return run_admix_pipeline(config)
    elif config.command == "mito":
        return setup_mito_mapping(config)
    elif config.command == "example":
        return setup_example(config)

    return 1



#!/usr/bin/python
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Tool for building histogram of PCR duplicates to be used with 'preseq'.

This allows estimation of the library complexity, and potential further gains
from sequencing the library. Unlike the tools included in 'preseq', this tool
handles collapsed reads.

Preseq is located at http://smithlabresearch.org/software/preseq/
"""
import sys
import random
import argparse
import collections

import pysam

import paleomix.common.bamfiles as bamfiles


def get_template_length(record):
    """Returns the template length of the given record for paired or collapsed
    reads; for single-ended reads, None is returned."""
    if record.is_paired:
        return record.tlen
    elif record.qname.startswith("M_"):
        return record.alen
    return None


def process_records(records, counts):
    """Processes a set of records aligned to the same positon; dulpicates are
    inferred based on the template lengths of the records, using the sequence
    length for collapsed reads (where qname.startswith("M_")). Single-ended
    reads are assumed to represent a random sampling of reads for which the
    insert size is known.
    """
    alignments = collections.defaultdict(int)
    for record in records:
        if record.is_paired:
            if (not record.is_proper_pair) or record.is_read2:
                continue

        alignment = get_template_length(record)
        alignments[alignment] += 1

    if (None in alignments) and len(alignments) > 1:
        ambigious_count = alignments.pop(None)

        # PE reads are assummed to represent a random sample of PE / collapsed
        # reads, and distributed randomly across these to approximate this
        keys = tuple(alignments)
        for _ in xrange(ambigious_count):
            key = random.choice(keys)
            alignments[key] += 1

    for count in alignments.itervalues():
        counts[count] += 1


def parse_args(argv):
    prog = "paleomix duphist"
    usage = "%s sorted.bam > out.histogram" % (prog,)
    parser = argparse.ArgumentParser(prog=prog, usage=usage)
    parser.add_argument("bamfile", help="Sorted BAM file.")

    return parser.parse_args(argv)


def main(argv):
    """Main function; takes a list of arguments equivalent to sys.argv[1:]."""
    args = parse_args(argv)

    # Default filters, excepting that PCR duplicates are not filtered
    mask = bamfiles.EXCLUDED_FLAGS & ~bamfiles.BAM_PCR_DUPLICATE
    counts = collections.defaultdict(int)
    with pysam.Samfile(args.bamfile) as handle:
        for region in bamfiles.BAMRegionsIter(handle, exclude_flags=mask):
            for (_, records) in region:
                process_records(records, counts)

    for (key, count) in sorted(counts.iteritems()):
        print "%i\t%i" % (key, count)


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python
#
# Copyright (c) 2013 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import optparse

import paleomix

import paleomix.tools.phylo_pipeline.parts.genotype as genotype
import paleomix.tools.phylo_pipeline.parts.msa as msa
import paleomix.tools.phylo_pipeline.parts.paml as paml
import paleomix.tools.phylo_pipeline.parts.phylo as phylo
import paleomix.common.console as console

from paleomix.config import \
     ConfigError, \
     PerHostValue, \
     PerHostConfig


_DESCRIPTION = \
  "Commands:\n" \
  "  -- %prog help            -- Display this message.\n" \
  "  -- %prog example [...]   -- Copy example project to folder.\n" \
  "  -- %prog makefile        -- Print makefile template.\n" \
  "  -- %prog genotype [...]  -- Carry out genotyping according to makefile.\n" \
  "  -- %prog msa [...]       -- Carry out multiple sequence alignment.\n" \
  "  -- %prog phylogeny [...] -- Carry out phylogenetic inference.\n"


_COMMANDS = {
    "mkfile"          : True,
    "makefile"        : True,
    "genotype"        : genotype.chain,
    "genotyping"      : genotype.chain,
    "msa"             : msa.chain,
    "paml:codeml"     : paml.chain_codeml,
    "phylogeny:examl" : phylo.chain_examl,
    }


class CustomHelpFormatter(optparse.IndentedHelpFormatter):
    def format_description(self, description):
        return description or ""


def select_commands(chain):
    commands = []
    for command in chain.split("+"):
        command_key  = command.strip().lower()
        command_func = None

        if command in _COMMANDS:
            command_func = _COMMANDS[command_key]
        elif len(command) >= 3:
            for (key, value) in _COMMANDS.iteritems():
                if key.startswith(command):
                    command_key  = key
                    command_func = value
                    break

        commands.append((command_key, command_func))

    return commands


def _run_config_parser(argv):
    per_host_cfg = PerHostConfig("phylo_pipeline")
    usage_str = "paleomix phylo_pipeline <command> [options] [makefiles]"
    version_str = "paleomix phylo_pipeline v%s" % (paleomix.__version__,)
    parser = optparse.OptionParser(usage=usage_str,
                                   version=version_str)

    parser.formatter = CustomHelpFormatter()
    parser.formatter.set_parser(parser)
    parser.description = _DESCRIPTION

    paleomix.ui.add_optiongroup(parser,
                                ui_default=PerHostValue("running"),
                                color_default=PerHostValue("on"))
    paleomix.logger.add_optiongroup(parser, default = PerHostValue("warning"))

    group  = optparse.OptionGroup(parser, "Scheduling")
    group.add_option("--samtools-max-threads",  default = PerHostValue(1), type = int,
                     help = "Maximum number of threads to use when genotyping or building pileups [%default]")
    group.add_option("--examl-max-threads",  default = PerHostValue(1), type = int,
                     help = "Maximum number of threads to use for each instance of ExaML [%default]")
    group.add_option("--max-threads",        default = per_host_cfg.max_threads, type = int,
                     help = "Maximum number of threads to use in total [%default]")
    group.add_option("--dry-run",            default = False, action="store_true",
                     help = "If passed, only a dry-run in performed, the dependency tree is printed, "
                            "and no tasks are executed.")
    parser.add_option_group(group)

    group  = optparse.OptionGroup(parser, "Required paths")
    group.add_option("--temp-root",    default = per_host_cfg.temp_root,
                     help = "Location for temporary files and folders [%default]")
    group.add_option("--samples-root", default = PerHostValue("./data/samples", is_path = True),
                     help = "Location of BAM files for each sample [%default]")
    group.add_option("--regions-root", default = PerHostValue("./data/regions", is_path = True),
                     help = "Location of BED files containing regions of interest [%default]")
    group.add_option("--prefix-root",  default = PerHostValue("./data/prefixes", is_path = True),
                     help = "Location of prefixes (FASTAs) [%default]")
    group.add_option("--refseq-root",  default = PerHostValue("./data/refseqs", is_path = True),
                     help = "Location of reference sequences (FASTAs) [%default]")
    group.add_option("--destination",  default = "./results",
                     help = "The destination folder for result files [%default]")
    parser.add_option_group(group)

    group = optparse.OptionGroup(parser, "Files and executables")
    group.add_option("--list-input-files", action="store_true", default=False,
                     help="List all input files used by pipeline for the "
                          "makefile(s), excluding any generated by the "
                          "pipeline itself.")
    group.add_option("--list-output-files", action="store_true", default=False,
                     help="List all output files generated by pipeline for "
                          "the makefile(s).")
    group.add_option("--list-executables", action="store_true", default=False,
                     help="List all executables required by the pipeline, "
                          "with version requirements (if any).")
    parser.add_option_group(group)

    parser.add_option("--to-dot-file", dest="dot_file",
                      help="Write dependency tree to the specified dot-file.")

    return per_host_cfg.parse_args(parser, argv)


def parse_config(argv):
    options, args = _run_config_parser(argv)
    paleomix.ui.set_ui_colors(options.ui_colors)

    if args and args[0] in ("example", "examples"):
        return options, args
    elif (len(args) < 2) and (args != ["mkfile"] and args != ["makefile"]):
        description = _DESCRIPTION.replace("%prog", "phylo_pipeline").strip()
        console.print_info("Phylogeny Pipeline v%s\n" % (paleomix.__version__,))
        console.print_info(description)
        return options, args

    commands = select_commands(args[0] if args else ())
    if any((func is None) for (_, func) in commands):
        unknown_commands = ", ".join(repr(key) for (key, func) in commands if func is None)
        raise ConfigError("Unknown analysis step(s): %s" % (unknown_commands,))

    return options, args
